'''
bam_vs_bed.py - count context that reads map to
======================================================

:Tags: Genomics NGS Intervals BAM BED Counting

Purpose
-------

This script takes as input a :term:`BAM` file from an RNA-seq or
similar experiment and a :term:`bed` formatted file. The :term:`bed`
formatted file needs at least four columns. The fourth (name) column
is used to group counts.

The script counts the number of alignments overlapping in the first
input file that overlap each feature in the second file. Annotations
in the :term:`bed` file can be overlapping - they are counted
independently.

This scripts requires bedtools_ to be installed.

Options
-------

-a, --bam-file / -b, --bed-file
    These are the input files. They can also be provided as provided as
    positional arguements, with the bam file being first and the (gziped
    or uncompressed) bed file coming second

-m, --min-overlap
    Using this option will only count reads if they overlap with a bed entry
    by a certain minimum fraction of the read.

Example
-------

Example::

   python bam_vs_bed.py in.bam in.bed.gz

Usage
-----

Type::

   cgat bam_vs_bed BAM BED [OPTIONS]
   cgat bam_vs_bed --bam-file=BAM --bed-file=BED [OPTIONS]

where BAM is either a bam or bed file and BED is a bed file.

Type::

   cgat bam_vs_bed --help

for command line help.

Command line options
--------------------

'''

import sys
import collections
import itertools
import subprocess
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import pysam
import CGAT.Bed as Bed


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--min-overlap", dest="min_overlap",
                      type="float",
                      help="minimum overlap [%default]")

    parser.add_option("-a", "--bam-file", dest="filename_bam",
                      metavar="bam", type="string",
                      help="bam-file to use (required) [%default]")

    parser.add_option("-b", "--bed-file", dest="filename_bed",
                      metavar="bed", type="string",
                      help="bed-file to use (required) [%default]")

    parser.add_option(
        "-s", "--sort-bed", dest="sort_bed",
        action="store_true",
        help="sort the bed file by chromosomal location before "
        "processing. "
        "[%default]")

    parser.add_option(
        "--assume-sorted", dest="sort_bed",
        action="store_false",
        help="assume that the bed-file is sorted by chromosomal location. "
        "[%default]")

    parser.add_option(
        "--split-intervals", dest="split_intervals",
        action="store_true",
        help="treat split BAM intervals, for example spliced intervals, "
        "as separate intervals. Note that a single alignment might be "
        "counted several times as a result. "
        "[%default]")

    parser.set_defaults(
        min_overlap=0.5,
        filename_bam=None,
        filename_bed=None,
        sort_bed=True,
        split_intervals=False,
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)

    filename_bam = options.filename_bam
    filename_bed = options.filename_bed

    if filename_bam is None and filename_bed is None:
        if len(args) != 2:
            raise ValueError(
                "please supply a bam and a bed file or two bed-files.")

        filename_bam, filename_bed = args

    if filename_bed is None:
        raise ValueError("please supply a bed file to compare to.")

    if filename_bam is None:
        raise ValueError("please supply a bam file to compare with.")

    E.info("intersecting the two files")

    min_overlap = options.min_overlap

    options.stdout.write("category\talignments\n")

    # get number of columns of reference bed file
    for bed in Bed.iterator(IOTools.openFile(filename_bed)):
        ncolumns_bed = bed.columns
        break
    E.info("assuming %s is bed%i format" % (filename_bed, ncolumns_bed))

    if ncolumns_bed < 4:
        raise ValueError("please supply a name attribute in the bed file")

    # get information about
    if filename_bam.endswith(".bam"):
        format = "-abam"
        samfile = pysam.AlignmentFile(filename_bam, "rb")
        total = samfile.mapped
        # latest bedtools uses bed12 format when bam is input
        ncolumns_bam = 12
        # count per read
        sort_key = lambda x: x.name
    else:
        format = "-a"
        total = IOTools.getNumLines(filename_bam)
        # get bed format
        ncolumns_bam = 0
        for bed in Bed.iterator(IOTools.openFile(filename_bam)):
            ncolumns_bam = bed.columns
            break

        if ncolumns_bam > 0:
            E.info("assuming %s is bed%i fomat" % (filename_bam, ncolumns_bam))
            if ncolumns_bam == 3:
                # count per interval
                sort_key = lambda x: (x.contig, x.start, x.end)
            else:
                # count per interval category
                sort_key = lambda x: x.name

    # use fields for bam/bed file (regions to count with)
    data_fields = [
        "contig", "start", "end", "name",
        "score", "strand", "thickstart", "thickend", "rgb",
        "blockcount", "blockstarts", "blockends"][:ncolumns_bam]

    # add fields for second bed (regions to count in)
    data_fields.extend([
        "contig2", "start2", "end2", "name2",
        "score2", "strand2", "thickstart2", "thickend2", "rgb2",
        "blockcount2", "blockstarts2", "blockends2"][:ncolumns_bed])

    # add bases overlap
    data_fields.append("bases_overlap")

    data = collections.namedtuple("data", data_fields)

    options.stdout.write("total\t%i\n" % total)

    if total == 0:
        E.warn("no data in %s" % filename_bam)
        return

    # SNS: sorting optional, off by default
    if options.sort_bed:
        bedcmd = "<( gunzip < %s | sort -k1,1 -k2,2n)" % filename_bed
    else:
        bedcmd = filename_bed

    if options.split_intervals:
        split = "-split"
    else:
        split = ""

    # IMS: newer versions of intersectBed have a very high memory
    #      requirement unless passed sorted bed files.
    statement = """bedtools intersect %(format)s %(filename_bam)s
    -b %(bedcmd)s
    %(split)s
    -sorted -bed -wo -f %(min_overlap)f""" % locals()

    E.info("starting counting process: %s" % statement)
    proc = E.run(statement,
                 return_popen=True,
                 stdout=subprocess.PIPE)

    E.info("counting")
    counts_per_alignment = collections.defaultdict(int)
    take_columns = len(data._fields)

    def iterate(infile):
        for line in infile:
            if not line.strip():
                continue
            yield data._make(line[:-1].split()[:take_columns])

    for read, overlaps in itertools.groupby(
            iterate(IOTools.force_str(proc.stdout)), key=sort_key):
        annotations = [x.name2 for x in overlaps]
        for anno in annotations:
            counts_per_alignment[anno] += 1

    for key, counts in sorted(counts_per_alignment.items()):
        options.stdout.write("%s\t%i\n" % (key, counts))

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''
R.py - Intergrate python and R
===========================================================

:Tags: Python

Purpose
-------

This module file contains code to allow integration of python and
R:

   - Redefined R class which enables saving R history.

Usage
-----
'''

from rpy2.robjects import R as RBase


class R_with_History(RBase):
    ''' Redefine the RBase class to allow history to be saved'''
    _instance = None

    def __new__(cls):
        c = RBase.__new__(cls)
        cls._instance = c
        c._history = []
        return cls._instance

    def __call__(self, string):
        self._history.append(string)
        RBase.__call__(self, string)

    def loadImage(self, imageFile):
        self["load.image"](file=imageFile)

    def saveImage(self, imageFile):
        self["save.image"](file=imageFile)

    def saveHistory(self, historyFile, append=False):
        ''' save history '''

        filetype = "w"
        if append:
            filetype = "a"

        with open(historyFile, filetype) as outf:
            outf.write("\n".join(self._history) + "\n")



"""
BamTools - Utilities for working with BAM files
===============================================

This module brings together convenience function for working
with :term:`bam` formatted files.

"""

import numpy
import pysam


def isPaired(bamfile, alignments=1000):
    '''check if a `bamfile` contains paired end data

    The method reads at most the first *alignments* and returns
    True if any of the alignments are paired.
    '''

    samfile = pysam.AlignmentFile(bamfile)
    n = 0
    for read in samfile:
        if read.is_paired:
            break
        n += 1
        if n == alignments:
            break

    samfile.close()

    return n != alignments


def estimateInsertSizeDistribution(bamfile,
                                   alignments=10000,
                                   n=10,
                                   method="picard",
                                   similarity_threshold=1.0,
                                   max_chunks=1000):
    '''estimate insert size from a subset of alignments in a bam file.

    Several methods are implemented.

    picard
        The method works analogous to picard by restricting the estimates
        to a core distribution. The core distribution is defined as all
        values that lie within n-times the median absolute deviation of
        the full data set.
    convergence
        The method works similar to ``picard``, but continues reading
        `alignments` until the mean and standard deviation stabilize.
        The values returned are the median mean and median standard
        deviation encountered.

    The method `convergence` is suited to RNA-seq data, as insert sizes
    fluctuate siginificantly depending on the current region
    being looked at.

    Only mapped and proper pairs are considered in the computation.

    Returns
    -------
    mean : float
       Mean of insert sizes.
    stddev : float
       Standard deviation of insert sizes.
    npairs : int
       Number of read pairs used for the estimation
    method : string
       Estimation method
    similarity_threshold : float
       Similarity threshold to apply.
    max_chunks : int
       Maximum number of chunks of size `alignments` to be used
       in the convergence method.

    '''

    assert isPaired(bamfile), \
        'can only estimate insert size from' \
        'paired bam files'

    samfile = pysam.AlignmentFile(bamfile)

    def get_core_distribution(inserts, n):
        # compute median absolute deviation
        raw_median = numpy.median(inserts)
        raw_median_dev = numpy.median(numpy.absolute(inserts - raw_median))

        # set thresholds
        threshold_min = max(0, raw_median - n * raw_median_dev)
        threshold_max = raw_median + n * raw_median_dev

        # define core distribution
        return inserts[numpy.logical_and(inserts >= threshold_min,
                                         inserts <= threshold_max)]

    if method == "picard":

        # only get first read in pair to avoid double counting
        inserts = numpy.array(
            [read.template_length for read in samfile.head(n=alignments)
             if read.is_proper_pair
             and not read.is_unmapped
             and not read.mate_is_unmapped
             and not read.is_read1
             and not read.is_duplicate
             and read.template_length > 0])
        core = get_core_distribution(inserts, n)

        return numpy.mean(core), numpy.std(core), len(inserts)

    elif method == "convergence":

        means, stds, counts = [], [], []
        last_mean = 0
        iteration = 0
        while iteration < max_chunks:

            inserts = numpy.array(
                [read.template_length for read in samfile.head(
                    n=alignments,
                    multiple_iterators=False)
                 if read.is_proper_pair
                 and not read.is_unmapped
                 and not read.mate_is_unmapped
                 and not read.is_read1
                 and not read.is_duplicate
                 and read.template_length > 0])
            core = get_core_distribution(inserts, n)
            means.append(numpy.mean(core))
            stds.append(numpy.std(core))
            counts.append(len(inserts))
            mean_core = get_core_distribution(numpy.array(means), 2)
            mm = numpy.mean(mean_core)
            if abs(mm - last_mean) < similarity_threshold:
                break
            last_mean = mm

        return numpy.median(means), numpy.median(stds), sum(counts)
    else:
        raise ValueError("unknown method '%s'" % method)


def estimateTagSize(bamfile,
                    alignments=10,
                    multiple="error"):
    '''estimate tag/read size from first alignments in file.

    Arguments
    ---------
    bamfile : string
       Filename of :term:`bam` formatted file
    alignments : int
       Number of alignments to inspect
    multiple : string
       How to deal if there are multiple tag sizes present.
       ``error`` will raise a warning, ``mean`` will return the
       mean of the read lengths found. ``uniq`` will return a
       unique list of read sizes found. ``all`` will return all
       read sizes encountered.

    Returns
    -------
    size : int
       The read size (actual, mean or list of read sizes)

    Raises
    ------
    ValueError
       If there are multiple tag sizes present and `multiple` is set to
       `error`.

    '''
    samfile = pysam.AlignmentFile(bamfile)
    sizes = [read.rlen for read in samfile.head(alignments)]
    mi, ma = min(sizes), max(sizes)

    if mi == 0 and ma == 0:
        sizes = [read.inferred_length for read in samfile.head(alignments)]
        # remove 0 sizes (unaligned reads?)
        sizes = [x for x in sizes if x > 0]
        mi, ma = min(sizes), max(sizes)

    if mi != ma:
        if multiple == "error":
            raise ValueError('multiple tag sizes in %s: %s' % (bamfile, sizes))
        elif multiple == "mean":
            mi = int(sum(sizes) / len(sizes))
        elif multiple == "uniq":
            mi = list(sorted(set(sizes)))
        elif multiple == "all":
            return sizes

    return mi


def getNumberOfAlignments(bamfile):
    '''return number of alignments in bamfile.
    '''
    samfile = pysam.AlignmentFile(bamfile)
    return samfile.mapped


def getNumReads(bamfile):
    '''count number of reads in bam file.

    This methods works through pysam.idxstats.

    Arguments
    ---------
    bamfile : string
        Filename of :term:`bam` formatted file. The file needs
        to be indexed.
    Returns
    -------
    nreads : int
        Number of reads
    '''

    lines = pysam.idxstats(bamfile).splitlines()

    try:
        nreads = sum(
            map(int, [x.split("\t")[2]
                      for x in lines if not x.startswith("#")]))

    except IndexError as msg:
        raise IndexError(
            "can't get number of reads from bamfile, msg=%s, data=%s" %
            (msg, lines))
    return nreads


def isStripped(bamfile):
    '''
    Check if the sequence is stripped in a bam file.
    '''
    bam = pysam.AlignmentFile(bamfile)
    read = bam.next()
    if read.seq is None:
        return True
    else:
        return False



'''diff_fasta.py - compare contents of two fasta files
===================================================

:Tags: Genomics Sequences FASTA Comparison

Purpose
-------

This script takes two sets of fasta sequences and matches the
identifiers. It then compares the sequences with the same identifiers
and, depending on the output options selected, outputs

   * which sequences are missing
   * which sequences are identical
   * which sequences are prefixes/suffixes of each other

An explanatory field is appended to output sequence identifiers.
An explanation of the different field values is provided in the log.

Options
-------
-s, --correct-gap-shift
   This option will correct shifts in alignment gaps between
   two sequences being compared

-1, --pattern1
   regular expression pattern to extract identifier from in sequence 1

-2, --pattern2
   regular expression pattern to extract identifier from in sequence 2

Depending on the option ``--output-section`` the following are output:

  diff
     identifiers of sequences that are different

  seqdiff
     identifiers of sequences that are different plus sequence

  missed
    identifiers of seqences that are missing from one set or the other

This script is of specialized interest and has been used
in the past to check if ENSEMBL gene models had been
correctly mapped into a database schema.

Usage
-----

Example::

   cat a.fasta | head

   >ENSACAP00000004922
   MRSRNQGGESSSSGKFSKSKPIINTGENQNLQEDAKKKNKSSRKEE ...
   >ENSACAP00000005213
   EEEEDESNNSYLYQPLNQDPDQGPAAVEETAPSTEPALDINERLQA ...
   >ENSACAP00000018122
   LIRSSSMFHIMKHGHYISRFGSKPGLKCIGMHENGIIFNNNPALWK ...

   python diff_fasta.py --output-section=missed --output-section=seqdiff a.fasta b.fasta

   cat diff.out

   # Legend:
   # seqs1:          number of sequences in set 1
   # seqs2:          number of sequences in set 2
   # same:           number of identical sequences
   # diff:           number of sequences with differences
   # nmissed1:       sequences in set 1 that are not found in set 2
   # nmissed2:       sequences in set 2 that are not found in set 1
   # Type of sequence differences
   # first:          only the first residue is different
   # last:           only the last residue is different
   # prefix:         one sequence is prefix of the other
   # selenocysteine: difference due to selenocysteines
   # masked:         difference due to masked residues
   # fixed:          fixed differences
   # other:          other differences


Type::

   python diff_fasta.py --help

for command line help.

Command line options
--------------------

'''
import sys
import re
import CGAT.FastaIterator as FastaIterator
import CGAT.Experiment as E
import CGAT.IOTools as IOTools


def MapIdentifiers(seqs, pattern):

    rx = re.compile(pattern)

    for k, s in list(seqs.items()):
        try:
            nk = rx.search(k).groups()[0]
        except AttributeError:
            raise ValueError(
                "identifier can not be parsed from '%s' "
                "pattern='%s'" % (k, pattern))

        del seqs[k]
        seqs[nk] = s


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "-s", "--correct-gap-shift", dest="correct_shift",
        action="store_true",
        help="correct gap length shifts in alignments. "
        "Requires alignlib_lite.py [%default]")

    parser.add_option(
        "-1", "--pattern1", dest="pattern1", type="string",
        help="pattern to extract identifier from in identifiers1. "
        "[%default]")

    parser.add_option(
        "-2", "--pattern2", dest="pattern2", type="string",
        help="pattern to extract identifier from in identifiers2. "
        "[%default]")

    parser.add_option(
        "-o", "--output-section", dest="output", type="choice",
        action="append",
        choices=("diff", "missed", "seqdiff"),
        help="what to output [%default]")

    parser.set_defaults(correct_shift=False,
                        pattern1="(\S+)",
                        pattern2="(\S+)",
                        output=[])

    (options, args) = E.Start(parser)

    if len(args) != 2:
        raise ValueError("two files needed to compare.")

    if options.correct_shift:
        try:
            import alignlib_lite
        except ImportError:
            raise ImportError(
                "option --correct-shift requires alignlib_lite.py_ "
                "but alignlib not found")

    seqs1 = dict([
        (x.title, x.sequence) for x in FastaIterator.iterate(
            IOTools.openFile(args[0], "r"))])
    seqs2 = dict([
        (x.title, x.sequence) for x in FastaIterator.iterate(
            IOTools.openFile(args[1], "r"))])

    if not seqs1:
        raise ValueError("first file %s is empty." % (args[0]))
    if not seqs2:
        raise ValueError("second file %s is empty." % (args[1]))

    MapIdentifiers(seqs1, options.pattern1)
    MapIdentifiers(seqs2, options.pattern2)

    nsame = 0
    nmissed1 = 0
    nmissed2 = 0
    ndiff = 0
    ndiff_first = 0
    ndiff_last = 0
    ndiff_prefix = 0
    ndiff_selenocysteine = 0
    ndiff_masked = 0
    nfixed = 0
    found2 = {}

    write_missed1 = "missed" in options.output
    write_missed2 = "missed" in options.output
    write_seqdiff = "seqdiff" in options.output
    write_diff = "diff" in options.output or write_seqdiff

    for k in sorted(seqs1):
        if k not in seqs2:
            nmissed1 += 1
            if write_missed1:
                options.stdout.write("---- %s ---- %s\n" % (k, "missed1"))
            continue

        found2[k] = 1

        s1 = seqs1[k].upper()
        s2 = seqs2[k].upper()
        m = min(len(s1), len(s2))

        if s1 == s2:
            nsame += 1
        else:
            status = "other"

            ndiff += 1

            if s1[1:] == s2[1:]:
                ndiff_first += 1
                status = "first"
            elif s1[:m] == s2[:m]:
                ndiff_prefix += 1
                status = "prefix"
            elif s1[:-1] == s2[:-1]:
                ndiff_last += 1
                status = "last"
            else:
                if len(s1) == len(s2):
                    # get all differences: the first and last residues
                    # can be different for peptide sequences when
                    # comparing my translations with ensembl peptides.
                    differences = []
                    for x in range(1, len(s1) - 1):
                        if s1[x] != s2[x]:
                            differences.append((s1[x], s2[x]))

                    l = len(differences)
                    # check for Selenocysteins
                    if len([x for x in differences if x[0] == "U" or x[1] == "U"]) == l:
                        ndiff_selenocysteine += 1
                        status = "selenocysteine"

                    # check for masked residues
                    elif len([x for x in differences if x[0] in "NX" or x[1] in "NX"]) == l:
                        ndiff_masked += 1
                        status = "masked"

            # correct for different gap lengths
            if options.correct_shift:

                map_a2b = alignlib_lite.py_makeAlignmentVector()

                a, b = 0, 0
                keep = False

                x = 0
                while x < m and not (a == len(s1) and b == len(s2)):
                    try:
                        if s1[a] != s2[b]:
                            while s1[a] == "N" and s2[b] != "N":
                                a += 1
                            while s1[a] != "N" and s2[b] == "N":
                                b += 1

                            if s1[a] != s2[b]:
                                break
                    except IndexError:
                        print("# index error for %s: x=%i, a=%i, b=%i, l1=%i, l2=%i" % (k, x, a, b, len(s1), len(s2)))
                        break

                    a += 1
                    b += 1
                    map_a2b.addPairExplicit(a, b, 0.0)
                    # check if we have reached the end:
                else:
                    keep = True
                    nfixed += 1
                    f = alignlib_lite.py_AlignmentFormatEmissions(map_a2b)
                    print("fix\t%s\t%s" % (k, str(f)))

                if not keep:
                    print("# warning: not fixable: %s" % k)

            if write_diff:
                options.stdout.write("---- %s ---- %s\n" % (k, status))

            if write_seqdiff:
                options.stdout.write("< %s\n> %s\n" % (seqs1[k], seqs2[k]))

    for k in sorted(list(seqs2.keys())):
        if k not in found2:
            nmissed2 += 1
            if write_missed2:
                options.stdout.write("---- %s ---- %s\n" % (k, "missed2"))

    options.stdlog.write("""# Legend:
""")

    E.info("seqs1=%i, seqs2=%i, same=%i, ndiff=%i, nmissed1=%i, nmissed2=%i" %
           (len(seqs1), len(seqs2), nsame, ndiff, nmissed1, nmissed2))

    E.info(
        "ndiff=%i: first=%i, last=%i, prefix=%i, selenocysteine=%i, masked=%i, fixed=%i, other=%i" %
        (ndiff, ndiff_first, ndiff_last, ndiff_prefix,
         ndiff_selenocysteine, ndiff_masked, nfixed,
         ndiff - ndiff_first - ndiff_last - ndiff_prefix -
         ndiff_selenocysteine - ndiff_masked - nfixed))

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''bam2bam.py - modify bam files
=============================

:Tags: Genomics NGS BAM Manipulation

Purpose
-------

This script reads a :term:`bam` formatted file from stdin, performs an
action (see methods below) then outputs a modified :term:`bam`
formatted file on stdout.

.. note::
   You need to redirect logging information to a file (via -L) or turn it off
   via -v 0 in order to get a valid sam/bam file.

Documentation
-------------

The script implements the following methods:

``set-nh``

   set the NH flag. Some tools (bowtie, bwa) do not set the NH flag.
   If set, this option will set the NH flag (for mapped reads).
   This option requires the bam/sam file to be sorted by read name.

``unset-unmapped_mapq``

   some tools set the mapping quality of unmapped reads. This
   causes a violation in the Picard tools.

``filter``

   remove alignments based on a variety of flags. The filtering method
   is determined by the ``--filter-method`` option. These may be
   ``unique``, ``non-unique``, ``mapped``, ``NM`` or ``CM``.  If
   ``unique`` is set, only uniquely mapping reads will be output. If
   ``non-unique`` is set then only multi-mapping reads will be
   output. This method first checks for the NH flag - if set, a unique
   match should have at most NH=1 hits.  If not set, the method checks
   for BWA flags. Currently it checks if X0 is set (X0=Number of best
   hits found by BWA).  If ``mapped`` is given, unmapped reads will be
   removed. If ``NM`` or ``CM`` is set, the alignment of reads in two
   sam files (input and reference) is compared and only reads with a
   lower number of mismatches in the input compared to the reference
   sam file will be kept. If ``CM`` is set, the colourspace mismatch
   tag (for ABI Solid reads) will be used to count differences to the
   reference sam file. By default, the ``NM`` (number of mismatches)
   tag is used. The tag that is used needs to present in both input
   sam file and the reference sam file. If ``unique`` is given this
   wil NOT remove any unmapped reads.  This can be achieved by
   providing the ``filter`` option twice, once each with ``mapped``
   and ``unique``.

   .. note::

      The filter methods can't currently combined with any of
      the other methods - this is work in progress.

``strip-sequence``

   remove the sequence from all reads in a bam-file. Note that
   stripping the sequence will also remove the quality scores.
   Stripping is not reversible if the read names are not unique.

``strip-quality``

   remove the quality scores from all reads in a bam-file.
   Stripping is not reversible if the read names are not unique.

``set-sequence``

   set the sequence and quality scores in the bam file to some dummy
   values ('A' for sequence, 'F' for quality which is a valid score in
   most fastq encodings. Necessary for some tools that can not work
   with bam-files without sequence.

``unstrip``

   add sequence and quality scores back to a bam file. Requires a
   :term:`fastq` formatted file with the sequences and quality scores
   to insert.

``unset-unmapped-mapq``

   sets the mapping quality of unmapped reads to 0.

``keep-first-base``

   keep only the first base of reads so that read counting tools will
   only consider the first base in the counts

``downsample-single``

   generates a downsampled :term:`bam` file by randomly subsampling
   reads from a single ended :term:`bam` file. The downsmpling
   retains multimapping reads. The use of this requires downsampling
   parameter to be set and optionally randomseed.

``downsample-paired``

   generates a downsampled :term:`bam` file by randomly subsampling
   reads from a paired ended :term:`bam` file. The downsampling
   retains multimapping reads. The use of this requires downsampling
   parameter to be set and optionally randomseed.


By default, the script works from stdin and outputs to stdout.
If the ``--inplace option`` is given, the script will modify
bam-files given as command line arguments in-place - a temporary
copy will be written to :file:`/tmp` and once completed copied
over the original file. The script will automatically re-index
the modified files.

Usage
-----

For example::

   cgat bam2bam --method=filter --filter-method=mapped < in.bam > out.bam

will remove all unmapped reads from the bam-file.

To remove unmapped reads from multiple bam-files, try::

   cgat bam2bam --inplace --method=filter --filter-method=mapped *.bam

example for running downsample::

   cgat bam2bam --method=downsample-paired --downsample=30000
        --randomseed=1 -L out.log < Paired.bam > out.bam

Type::

   cgat bam2bam --help

for command line help.

Command line options
--------------------

'''

import os
import sys
import tempfile
import shutil
import random
import pysam
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import itertools

try:
    import pyximport
    pyximport.install(build_in_temp=False)
    import _bam2bam
except ImportError:
    import CGAT.scripts._bam2bam as _bam2bam


class SubsetBam(object):

    ''' base class for performing downsampling on single and
    paired bam file

    A dictionary of the read names is made and then an
    array matching the numbers of reads in the dictionary will be
    created. This array contains 1s that match the number of reads
    to be downsampled to and 0s to fill out the rest of the array.

    The read is kept if the value in the dictionary matches 1. The
    read is yielded and iterated over to produce the output bam.

    The script will handle multimapping reads.
    '''

    def __init__(self, pysam_in, downsample, paired_end=None,
                 single_end=None, random_seed=None):

        self.pysam_in1, self.pysam_in2 = itertools.tee(pysam_in)
        self.downsample = downsample
        self.paired_end = paired_end
        self.single_end = single_end
        self.random_seed = random_seed

    def list_of_reads(self, paired=None):

        '''
        This will create a dictionary of uniqe reads in the bam
        '''
        read_list = []

        for read in self.pysam_in1:
            if paired is True:
                if read.is_proper_pair:
                    read_list.append(read.qname)
            else:
                read_list.append(read.qname)

        return sorted(set(read_list))

    def downsample_paired(self):

        '''
        This function will downsample a paired bam file.
        It will retain multimapping reads if they have not been
        pre-filtered
        '''
        if self.random_seed is not None:
            random.seed(self.random_seed)

        collect_list = self.list_of_reads(paired=True)
        read_list = random.sample(collect_list, self.downsample)

        if self.downsample == len(collect_list):
            E.warn('''The downsample reads is equal to the
            number of unique reads''')
            for read in self.pysam_in2:
                yield read

        # yield read if it is in read_list, if the read is multimapped
        # then all multimapping reads will be yielded
        else:
            E.warn('''Multimaping reads have been detected and these will
            be output to the final bam file''')
            for read in self.pysam_in2:
                if read.qname in read_list:
                    yield read

    def downsample_single(self):

        '''
        This function will downsample a single bam file.
        It will retain multimapping reads if not pre-filtered
        '''
        if self.random_seed is not None:
            random.seed(self.random_seed)

        collect_list = self.list_of_reads(paired=False)
        read_list = random.sample(collect_list, self.downsample)

        if self.downsample == len(collect_list):
            E.warn('''The downsample reads is equal to the
            number of unique reads''')
            for read in self.pysam_in2:
                yield read

        # yield read if it is in read_list, if the reads is multimapped
        # then all multimapping reads will be yielded
        else:
            E.warn('''Multimaping reads have been detected and these will
            be output to the final bam file''')
            for read in self.pysam_in2:
                if read.qname in read_list:
                    yield read


def main(argv=None):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv:
        argv = sys.argv

    # setup command line parser
    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-m", "--methods", dest="methods", type="choice",
                      action="append",
                      choices=("filter",
                               "keep-first-base",
                               "set-nh",
                               "set-sequence",
                               "strip-sequence",
                               "strip-quality",
                               "unstrip",
                               "unset-unmapped-mapq",
                               "downsample-single",
                               "downsample-paired"),
                      help="methods to apply [%default]")

    parser.add_option("--strip-method", dest="strip_method", type="choice",
                      choices=("all", "match"),
                      help="define which sequences/qualities to strip. "
                      "match means that stripping only applies to entries "
                      "without mismatches (requires NM tag to be present). "
                      "[%default]")

    parser.add_option("--filter-method", dest="filter_methods",
                      action="append", type="choice",
                      choices=('NM', 'CM', 'mapped', 'unique', "non-unique"),
                      help="filter method to apply to remove alignments "
                      "from a bam file. Multiple methods can be supplied "
                      "[%default]")

    parser.add_option("--reference-bam-file", dest="reference_bam",
                      type="string",
                      help="bam-file to filter with [%default]")

    parser.add_option("--force-output", dest="force", action="store_true",
                      help="force processing. Some methods such "
                      "as strip/unstrip will stop processing if "
                      "they think it not necessary "
                      "[%default]")

    parser.add_option("--output-sam", dest="output_sam", action="store_true",
                      help="output in sam format [%default]")

    parser.add_option("--inplace", dest="inplace", action="store_true",
                      help="modify bam files in-place. Bam files need "
                      "to be given "
                      "as arguments. Temporary bam files are written "
                      "to /tmp [%default]")

    parser.add_option(
        "--first-fastq-file", "-1", dest="fastq_pair1", type="string",
        help="fastq file with read information for first "
        "in pair or unpaired. Used for unstripping sequence "
        "and quality scores [%default]")

    parser.add_option(
        "--second-fastq-file", "-2", dest="fastq_pair2", type="string",
        help="fastq file with read information for second "
        "in pair. Used for unstripping sequence "
        "and quality scores  [%default]")

    parser.add_option(
        "--downsample", dest="downsample",
        type="int",
        help="Number of reads to downsample to")

    parser.set_defaults(
        methods=[],
        output_sam=False,
        reference_bam=None,
        filter_methods=[],
        strip_method="all",
        force=False,
        inplace=False,
        fastq_pair1=None,
        fastq_pair2=None,
        downsample=None,
        random_seed=None
    )

    # add common options (-h/--help, ...) and parse command line
    (options, args) = E.Start(parser, argv=argv)
    # random.seed(options.random_seed)
    bamfiles = []

    if options.stdin != sys.stdin:
        from_stdin = True
        bamfiles.append(options.stdin.name)
    else:
        from_stdin = False
        
    if options.inplace:
        bamfiles.extend(args)
        if len(bamfiles) == 0:
            raise ValueError(
                "please one or more bam-files as command line arguments")

        if "-" in bamfiles:
            raise ValueError(
                "can not read from stdin if ``--inplace`` is selected")

    if len(bamfiles) == 0:
        bamfiles = ["-"]

    to_stdout = False

    for bamfile in bamfiles:

        E.info('processing %s' % bamfile)

        if os.path.islink(bamfile):
            E.warn('ignoring link %s' % bamfile)
            continue

        if IOTools.isEmpty(bamfile):
            E.warn('ignoring empty file %s' % bamfile)
            continue

        # reading bam from stdin does not work with only the "r" tag
        pysam_in = pysam.AlignmentFile(bamfile, "rb")
        if bamfile == "-" or (from_stdin and bamfile == options.stdin.name):
            to_stdout = True
            if options.output_sam:
                pysam_out = pysam.AlignmentFile("-", "wh", template=pysam_in)
            else:
                pysam_out = pysam.AlignmentFile("-", "wb", template=pysam_in)
        else:
            if IOTools.isEmpty(bamfile):
                E.warn('skipping empty file %s' % bamfile)
                continue
            tmpfile = tempfile.NamedTemporaryFile(delete=False, prefix="ctmp")
            tmpfile.close()

            E.debug("writing temporary bam-file to %s" % tmpfile.name)
            pysam_out = pysam.AlignmentFile(tmpfile.name,
                                            "wb",
                                            template=pysam_in)

        if "filter" in options.methods:

            remove_mismatches, colour_mismatches = False, False

            if "NM" in options.filter_methods:
                remove_mismatches = True

            elif "CM" in options.filter_methods:
                remove_mismatches = True
                colour_mismatches = True

            if remove_mismatches:
                if not options.reference_bam:
                    raise ValueError(
                        "requiring reference bam file for removing by "
                        "mismatches")

                pysam_ref = pysam.AlignmentFile(options.reference_bam, "rb")
            else:
                pysam_ref = None

            # filter and flags are the opposite way around
            c = _bam2bam.filter_bam(
                pysam_in, pysam_out, pysam_ref,
                remove_nonunique="unique" in options.filter_methods,
                remove_unique="non-unique" in options.filter_methods,
                remove_contigs=None,
                remove_unmapped="mapped" in options.filter_methods,
                remove_mismatches=remove_mismatches,
                colour_mismatches=colour_mismatches)

            if pysam_ref:
                pysam_ref.close()

            # do not write to stdlog in the middle of a SAM/BAM stdout stream.
            if options.stdlog != options.stdout:
                E.info("category\tcounts\n%s\n" % c.asTable())
        else:

            # set up the modifying iterators
            it = pysam_in.fetch(until_eof=True)

            # function to check if processing should start
            pre_check_f = lambda x: None

            if "unset-unmapped-mapq" in options.methods:
                def unset_unmapped_mapq(i):
                    for read in i:
                        if read.is_unmapped:
                            read.mapq = 0
                        yield read
                it = unset_unmapped_mapq(it)

            if "set-sequence" in options.methods:
                def set_sequence(i):
                    for read in i:
                        # can't get at length of unmapped reads
                        if read.is_unmapped:
                            read.seq = "A"
                            read.qual = "F"
                        else:
                            read.seq = "A" * read.inferred_length
                            read.qual = "F" * read.inferred_length

                        yield read
                it = set_sequence(it)

            if "strip-sequence" in options.methods or "strip-quality" in \
               options.methods:
                def strip_sequence(i):
                    for read in i:
                        read.seq = None
                        yield read

                def check_sequence(reads):
                    if reads[0].seq is None:
                        return 'no sequence present'
                    return None

                def strip_quality(i):
                    for read in i:
                        read.qual = None
                        yield read

                def check_quality(reads):
                    if reads[0].qual is None:
                        return 'no quality information present'
                    return None

                def strip_match(i):
                    for read in i:
                        try:
                            nm = read.opt('NM')
                        except KeyError:
                            nm = 1
                        if nm == 0:
                            read.seq = None
                        yield read

                if options.strip_method == "all":
                    if "strip-sequence" in options.methods:
                        it = strip_sequence(it)
                        pre_check_f = check_sequence
                    elif "strip-quality" in options.methods:
                        it = strip_quality(it)
                        pre_check_f = check_quality
                elif options.strip_method == "match":
                    it = strip_match(it)

            if "unstrip" in options.methods:
                def buildReadDictionary(filename):
                    if not os.path.exists(filename):
                        raise OSError("file not found: %s" % filename)
                    fastqfile = pysam.FastxFile(filename)
                    fastq2sequence = {}
                    for x in fastqfile:
                        if x.name in fastq2sequence:
                            raise ValueError(
                                "read %s duplicate - can not unstrip" % x.name)

                        fastq2sequence[x.name] = (x.sequence, x.quality)
                    return fastq2sequence

                if not options.fastq_pair1:
                    raise ValueError(
                        "please supply fastq file(s) for unstripping")
                fastq2sequence1 = buildReadDictionary(options.fastq_pair1)
                if options.fastq_pair2:
                    fastq2sequence2 = buildReadDictionary(options.fastq_pair2)

                def unstrip_unpaired(i):
                    for read in i:
                        read.seq, read.qual = fastq2sequence1[read.qname]
                        yield read

                def unstrip_pair(i):
                    for read in i:
                        if read.is_read1:
                            read.seq, read.qual = fastq2sequence1[read.qname]
                        else:
                            read.seq, read.qual = fastq2sequence2[read.qname]
                        yield read

                if options.fastq_pair2:
                    it = unstrip_pair(it)
                else:
                    it = unstrip_unpaired(it)

            if "set-nh" in options.methods:
                it = _bam2bam.SetNH(it)

            # keep first base of reads by changing the cigarstring to
            # '1M' and, in reads mapping to the reverse strand,
            # changes the pos to aend - 1
            # Needs to be refactored to make it more general
            # (last base, midpoint, ..)
            if "keep_first_base" in options.methods:
                def keep_first_base(i):
                    for read in i:
                        if read.is_reverse:
                            read.pos = read.aend - 1
                            read.cigarstring = '1M'
                        elif not read.is_unmapped:
                            read.cigarstring = '1M'
                        yield read
                it = keep_first_base(it)

            # read first read and check if processing should continue
            # only possible when not working from stdin
            # Refactoring: use cache to also do a pre-check for
            # stdin input.
            if bamfile != "-":
                # get first read for checking pre-conditions
                first_reads = list(pysam_in.head(1))

                msg = pre_check_f(first_reads)
                if msg is not None:
                    if options.force:
                        E.warn('proccessing continues, though: %s' % msg)
                    else:
                        E.warn('processing not started: %s' % msg)
                        pysam_in.close()
                        pysam_out.close()
                        continue

            if "downsample-single" in options.methods:

                if not options.downsample:
                    raise ValueError("Please provide downsample size")

                else:
                    down = SubsetBam(pysam_in=it,
                                     downsample=options.downsample,
                                     paired_end=None,
                                     single_end=True,
                                     random_seed=options.random_seed)
                    it = down.downsample_single()

            if "downsample-paired" in options.methods:

                if not options.downsample:
                    raise ValueError("Please provide downsample size")

                else:
                    down = SubsetBam(pysam_in=it,
                                     downsample=options.downsample,
                                     paired_end=True,
                                     single_end=None,
                                     random_seed=options.random_seed)
                    it = down.downsample_paired()

            # continue processing till end
            for read in it:
                pysam_out.write(read)

        pysam_in.close()
        pysam_out.close()

        if options.inplace:
            # set date and file permissions according to original
            # Note: currently it will not update user and group.
            original = os.stat(bamfile)
            os.utime(tmpfile.name, (original.st_atime, original.st_mtime))
            os.chmod(tmpfile.name, original.st_mode)
            # move new file over original copy
            shutil.move(tmpfile.name, bamfile)
            # re-index
            pysam.index(bamfile)

    # write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""SequenceProperties.py - Computing metrics of nucleotide sequences
=================================================================

This module provides methods for extracting and reporting sequence
properties of nucleotide sequences such as the composition, length,
etc.

The classes provide the algorithms to provide the property. They will
store the latest result for output. Thus, processing is a two-step
procedure::

   from SequenceProperties import SequencePropertiesLength
   from SequenceProperties import SequencePropertiesNA

   counters = [SequencePropertiesLength(), SequencePropertiesNA()]

   # output column headers
   headers = sum(c.getHeaders() for c in counters]
   print "\t".join(headers)

   for sequence in sequences:
      # load sequence in each counter
      for c in counters:
          c.loadSequence(sequence)
      # output results
      print "\t".join(map(str, counters))

This design is useful to compute multiple properties while iterating
only once over an input file and output a single, multi-column table.

.. note::
    While useful and in working order, the design of the classes is
    cumbersome.

Reference
---------

"""

import re
import math
import hashlib
import base64
import itertools
import six

from CGAT import Genomics as Genomics

import Bio.Alphabet.IUPAC


class SequenceProperties(object):
    """Base class.

    This class is the base class for SequenceProperty objects. Derived
    classes need to overload most of its methods.
    """

    def __init__(self):

        self.mLength = 0
        self.mSeqType = "na"
        self.mSequence = ""

    def addProperties(self, other):
        """add properties."""
        self.mLength += other.mLength

    def updateProperties(self):
        pass

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""

        self.mSeqType = seqtype
        self.mSequence = re.sub("[ -.]", "", sequence).upper()
        self.mLength = len(self.mSequence)

    def __str__(self):
        return "\t".join(self.getFields())

    def getFields(self):
        self.updateProperties()
        return []

    def getHeaders(self):
        return []


class SequencePropertiesSequence(SequenceProperties):
    """Add properties: the actual sequence.

    sequence
       The sequence

    This class outputs the actual sequence supplied.
    """

    def __init__(self):
        SequenceProperties.__init__(self)

    def addProperties(self, other):
        """add properties."""
        SequenceProperties.addProperties(self, other)

    def updateProperties(self):
        SequenceProperties.updateProperties(self)

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)
        # self.mSequence = sequence

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        return fields + [self.mSequence, ]

    def getHeaders(self):
        headers = SequenceProperties.getHeaders(self)
        return headers + ["sequence"]


class SequencePropertiesHid(SequenceProperties):
    """Add properties: a hash of sequence

    hid
        Hash identifier of a sequence

    The hash is computed using the md5 algorithm and the resulting
    byte sequence is then translated into printable characters.
    """

    def __init__(self):
        SequenceProperties.__init__(self)
        self.mHid = ""

    def loadSequence(self, sequence, seqtype="na"):
        """load hid sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)

        # do the encryption
        h = hashlib.md5(six.b(sequence)).digest()
        # map to printable letters: hid has length 22, so the padded '=' are
        # truncated. You have to add them, if you ever want to decode,
        # but who would do such a thing :=)
        r = base64.encodestring(h)[0:22].decode("ascii")

        # finally substitute some characters:
        # '/' for '_', so we have legal file names
        # '[' for '+' and ']' for '=' for internet-applications
        hid = r.replace('/', '_').replace('+', '[').replace('=', ']')

        self.mHid = hid

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        return fields + [self.mHid, ]

    def getHeaders(self):
        headers = SequenceProperties.getHeaders(self)
        return headers + ["hid", ]


class SequencePropertiesLength(SequenceProperties):
    """Add properties: sequence length and number of codons

    length
        Sequence length
    ncodons
        Length in codons

    The number of codons is 0 for an amino-acid sequence.
    """

    def __init__(self):
        SequenceProperties.__init__(self)
        self.mLength = 0
        self.mNCodons = 0

    def addProperties(self, other):
        """add properties."""
        self.mLength += other.mLength
        self.mNCodons += other.mNCodons

    def updateProperties(self):
        SequenceProperties.updateProperties(self)

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)
        self.mLength = len(sequence)
        if self.mSeqType == "na":
            self.mNCodons = len(sequence) / 3
        else:
            self.mNCodons = 0

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        return fields + ["%i" % self.mLength,
                         "%i" % self.mNCodons]

    def getHeaders(self):
        headers = SequenceProperties.getHeaders(self)
        return headers + ["length", "ncodons"]


class SequencePropertiesNA(SequenceProperties):
    """Add properties: nucleotide composition

    nUnk
        Number of unknown residues
    nA, nC, nG, nT, nGC, nAT
        Nucleotide counts
    pA, pC, pG, pT, pGC, pAT
        Nucleotide frequencies

    """

    def __init__(self, reference_usage=[]):
        SequenceProperties.__init__(self)
        self.mCountsGC = 0
        self.mCountsAT = 0
        self.mCountsOthers = 0
        # counts of nucleotides
        self.mCountsNA = {}
        self.mAlphabet = Bio.Alphabet.IUPAC.unambiguous_dna.letters + "N"
        for x in self.mAlphabet:
            self.mCountsNA[x] = 0

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)
        self.mCountsGC += other.mCountsGC
        self.mCountsAT += other.mCountsAT
        self.mCountsOthers += other.mCountsOthers
        for na, count in list(other.mCountsNA.items()):
            self.mCountsNA[na] += count

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)
        # counts of nucleotides
        self.mCountsNA = {}
        for x in self.mAlphabet:
            self.mCountsNA[x] = 0

        for na in sequence.upper():
            if na in ('G', 'C'):
                self.mCountsGC += 1
            elif na in ('A', 'T'):
                self.mCountsAT += 1
            try:
                self.mCountsNA[na] += 1
            except KeyError:
                self.mCountsOthers += 1

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        # Counts
        fields.append("%i" % self.mCountsOthers)

        t = 0
        for x in self.mAlphabet:
            fields.append("%i" % self.mCountsNA[x])
            t += self.mCountsNA[x]
        fields.append("%i" % self.mCountsGC)
        fields.append("%i" % self.mCountsAT)
        # Percentages
        if t == 0:
            for x in self.mAlphabet:
                fields.append("na")
        else:
            for x in self.mAlphabet:
                fields.append("%f" % (float(self.mCountsNA[x]) / t))

        t2 = float(self.mCountsGC + self.mCountsAT)
        if t2 == 0:
            fields.append("na")
            fields.append("na")
        else:
            fields.append("%f" % (float(self.mCountsGC) / t2))
            fields.append("%f" % (float(self.mCountsAT) / t2))

        return fields

    def getHeaders(self):
        headers = SequenceProperties.getHeaders(self)
        # Counts
        headers.append("nUnk")
        for x in self.mAlphabet:
            headers.append("n%s" % x)
        headers.append("nGC")
        headers.append("nAT")
        # Percentages
        for x in self.mAlphabet:
            headers.append("p%s" % x)
        headers.append("pGC")
        headers.append("pAT")

        return headers


class SequencePropertiesDN(SequenceProperties):
    """Add Properties : dinucleotide counts

    nAA, nAC, ...
        Dinucleotide counts
    mCountsOthers
        Unknown dinucleotides
    """

    def __init__(self, reference_usage=[]):

        SequenceProperties.__init__(self)
        self.mCountsDinuc = {}
        self.mCountsOthers = 0
        self.mAlphabet = Bio.Alphabet.IUPAC.unambiguous_dna.letters
        for dinucleotide in itertools.product(self.mAlphabet, repeat=2):
            self.mCountsDinuc["".join(dinucleotide)] = 0

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)
        self.mCountsOthers += other.mCountsOthers
        for dn, count in list(other.mCountsDinuc.items()):
            self.mCountsDinuc[dn] += count

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)

        # IMS: generator rather than list
        # to save memory for might be a neater way of doing this.
        for dinuc in (sequence[x - 2:x]
                      for x in range(2, len(sequence) + 1, 1)):
            try:
                self.mCountsDinuc[dinuc] += 1
            except KeyError:
                self.mCountsOthers += 1

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        for x in itertools.product(self.mAlphabet, repeat=2):
            fields.append("%i" % self.mCountsDinuc["".join(x)])
        fields.append("%s" % self.mCountsOthers)
        return fields

    def getHeaders(self):

        headers = SequenceProperties.getHeaders(self)
        for dinucleotide in itertools.product(self.mAlphabet, repeat=2):
            headers.append("".join(dinucleotide))
        headers.append("mCountsOthers")
        return headers


class SequencePropertiesCpg(SequencePropertiesNA, SequencePropertiesDN):
    """Add Properties : CpG density and observed / expected.

    CpG_count
        Number of CpG in sequence
    CpG_density
        CpG density, number of CpG divided by 2 * sequence length
    CpG_ObsExp
        Ratio of observed to expected number of CpG. The latter is
        calculated as the product of nC * nG. The ratio is normalized
        by the sequence length.  Set to 0 if no ``C`` or ``G`` in
        sequence.

    """

    def __init__(self, reference_usage=[]):

        SequencePropertiesNA.__init__(self)
        SequencePropertiesDN.__init__(self)
        self.mCpG_density = 0
        self.mCpG_ObsExp = 0

    def addProperties(self, other):
        SequencePropertiesDN.addProperties(self, other)
        SequencePropertiesNA.addProperties(self, other)
        self.mCpG_density += other.mCpG_density
        self.mCpG_ObsExp += other.mCpG_ObsExp

    def loadSequence(self, sequence, seqtype="na"):
        SequencePropertiesNA.loadSequence(self, sequence, seqtype)
        SequencePropertiesDN.loadSequence(self, sequence, seqtype)

        if self.mLength > 0:
            self.mCpG_density = (float(self.mCountsDinuc["CG"]) /
                                 (float(self.mLength) / 2.0))
            if (float(self.mCountsNA["C"] * self.mCountsNA["G"])) / self.mLength > 0:
                self.mCpG_ObsExp = ((float(self.mCountsDinuc["CG"])) /
                                    ((float(self.mCountsNA["C"]) *
                                      float(self.mCountsNA["G"])) /
                                     float(self.mLength)))
            else:
                self.mCpG_ObsExp = 0.0
        else:
            self.mCpG_density = 0.0
            self.mCpG_ObsExp = 0.0

    def getFields(self):
        fields = ["%s" % self.mCountsDinuc["CG"]]
        fields.append("%6.4f" % self.mCpG_density)
        fields.append("%6.4f" % self.mCpG_ObsExp)
        return fields

    def getHeaders(self):
        headers = ["CpG_count"]
        headers.append("CpG_density")
        headers.append("CpG_ObsExp")
        return headers


class SequencePropertiesGaps(SequenceProperties):
    """Add Properties : number of gaps in a sequence

    Gaps are identified by unknown characters (``[XN]``)

    ngaps
         Number of gap characters in sequnce
    nseq_regions
         Number of ungapped regions
    ngap_regions
         Number of gapped regions
    """

    gap_chars = 'xXnN'

    def __init__(self,  gap_chars='xXnN', *args, **kwargs):
        SequenceProperties.__init__(self, *args, **kwargs)
        self.gap_chars = gap_chars

        self.ngaps = 0
        self.nseq_regions = 0
        self.ngap_regions = 0

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        return fields + ["%i" % self.ngaps,
                         "%i" % self.nseq_regions,
                         "%i" % self.ngap_regions]

    def getHeaders(self):

        headers = SequenceProperties.getHeaders(self)
        return headers + ["ngaps", "nseq_regions", "ngap_regions"]

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequenceProperties.loadSequence(self, sequence, seqtype)

        gap_chars = self.gap_chars
        ngaps = 0
        was_gap = not (sequence[0] in gap_chars)

        ngap_regions = 0
        nseq_regions = 0

        x = 0
        for x, c in enumerate(sequence):

            is_gap = c in gap_chars
            if is_gap:
                ngaps += 1
                if not was_gap:
                    ngap_regions += 1
            else:
                if was_gap:
                    nseq_regions += 1
            was_gap = is_gap

        self.ngaps = ngaps
        self.ngap_regions = ngap_regions
        self.nseq_regions = nseq_regions

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)

        self.ngaps += other.ngaps
        self.nseq_regions += other.nseq_regions
        self.ngap_regions += other.ngap_regions


class SequencePropertiesDegeneracy (SequencePropertiesLength):
    """Add properties : codon degeneracy

    nstops
        Number of stop codons
    nsites1d
        Number of non-degenerate sites
    nsites2d, nsites3d, nsites4d
        Number 2-fold, 3-fold, 4-fold degenerate sites
    ngc
        Number of positions containing either G or C
    ngc3
        Number of 3rd codon position containing G or C
    ngc3
        Number of non-degenerate 3rd codon position containing G or C
    n2gc3, n3gc3, n4gc3
        Number of 2-fold, 3-fold, 4-fold degenerate 3rd codon positions
        containing G or C
    pgc
        Percentage of positions containing either G or C
    pgc3
        Percentage of 3rd codon position containing G or C
    pgc3
        Percentage of non-degenerate 3rd codon position containing G or C
    p2gc3, p3gc3, p4gc3
        Percentage of 2-fold, 3-fold, 4-fold degenerate 3rd codon positions
        containing G or C

    The degeneracies for amino acids are::

       2: MW are non-degenerate.
       9: EDKNQHCYF are 2-fold degenerate.
       1: I is 3-fold degenerate
       5: VGATP are 4-fold degenerate.
       3: RLS are 2-fold and four-fold degenerate.
          Depending on the first two codons, the codons are counted
          as two or four-fold degenerate codons. This is encoded
          in the file Genomics.py.

    The number of degenerate sites is computed across all
    codon positions.

    """

    def __init__(self):

        SequencePropertiesLength.__init__(self)

        self.mNGC = 0
        self.mNSites1D, self.mNSites2D, self.mNSites3D, self.mNSites4D = (
            0, 0, 0, 0)
        self.mNGC3 = 0
        self.mN2DGC3 = 0
        self.mN3DGC3 = 0
        self.mN4DGC3 = 0

        self.mNStopCodons = 0

        # setup counting arrays
        # nucleotide counts for each position (is not a sum of the counts
        # per degenerate site, as the codon might be intelligible, e.g. GNN).
        self.mCounts = [{'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0},
                        {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0},
                        {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0}]

        # nucleotide counts for each position per degeneracy
        self.mCountsDegeneracy = []

        for x in (0, 1, 2):
            xx = []
            for y in range(5):
                yy = {}
                for z in Bio.Alphabet.IUPAC.extended_dna.letters:
                    yy[z] = 0
                xx.append(yy)
            self.mCountsDegeneracy.append(xx)

    def addProperties(self, other):

        SequencePropertiesLength.addProperties(self, other)
        self.mNStopCodons += other.mNStopCodons

        for x in (0, 1, 2):
            for y in range(5):
                for z in Bio.Alphabet.IUPAC.extended_dna.letters:
                    self.mCountsDegeneracy[x][y][
                        z] += other.mCountsDegeneracy[x][y][z]

        for x in (0, 1, 2):
            for y, count in list(other.mCounts[x].items()):
                self.mCounts[x][y] += count

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""

        SequencePropertiesLength.loadSequence(self, sequence, seqtype)
        if len(sequence) % 3:
            raise ValueError(
                '''sequence length is not a multiple of 3 (length=%i)''' %
                (len(sequence)))

        # uppercase all letters
        sequence = sequence.upper()

        self.mNStopCodons = 0

        # setup counting arrays
        # nucleotide counts for each position (is not a sum of the counts
        # per degenerate site, as the codon might be intelligible, e.g. GNN).
        self.mCounts = [{'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0},
                        {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0},
                        {'A': 0, 'C': 0, 'G': 0, 'T': 0, 'X': 0, 'N': 0}]

        # nucleotide counts for each position per degeneracy
        self.mCountsDegeneracy = []

        for x in (0, 1, 2):
            xx = []
            for y in range(5):
                yy = {}
                for z in Bio.Alphabet.IUPAC.extended_dna.letters:
                    yy[z] = 0
                xx.append(yy)
            self.mCountsDegeneracy.append(xx)

        # use generator rather than list to save memory
        for codon in (sequence[x:x + 3] for x in range(0, len(sequence), 3)):

            for x in (0, 1, 2):
                self.mCounts[x][codon[x]] += 1

            if Genomics.IsStopCodon(codon):
                self.mNStopCodons += 1
                continue

            try:
                aa, deg1, deg2, deg3 = Genomics.GetDegeneracy(codon)
                degrees = (deg1, deg2, deg3)
                for x in range(len(degrees)):
                    self.mCountsDegeneracy[x][degrees[x]][codon[x]] += 1

            except KeyError:
                pass

    def updateProperties(self):
        """update fields from counts."""

        SequencePropertiesLength.updateProperties(self)
        self.mNGC = 0
        self.mNSites1D, self.mNSites2D, self.mNSites3D, self.mNSites4D = (
            0,) * 4
        self.mNSitesD3 = 0

        for x in (0, 1, 2):
            self.mNGC += self.mCounts[x]['C'] + self.mCounts[x]['G']
            self.mNSites1D += sum(self.mCountsDegeneracy[x][1].values())
            self.mNSites2D += sum(self.mCountsDegeneracy[x][2].values())
            self.mNSites3D += sum(self.mCountsDegeneracy[x][3].values())
            self.mNSites4D += sum(self.mCountsDegeneracy[x][4].values())

        self.mNGC3 = self.mCounts[2]['C'] + self.mCounts[2]['G']
        self.mN2DGC3 = self.mCountsDegeneracy[2][2][
            'C'] + self.mCountsDegeneracy[2][2]['G']
        self.mN3DGC3 = self.mCountsDegeneracy[2][3][
            'C'] + self.mCountsDegeneracy[2][3]['G']
        self.mN4DGC3 = self.mCountsDegeneracy[2][4][
            'C'] + self.mCountsDegeneracy[2][4]['G']

        # count all degenerate sites at third codon position
        self.mNSitesD3 = sum(self.mCountsDegeneracy[2][2].values()) +\
            sum(self.mCountsDegeneracy[2][3].values()) +\
            sum(self.mCountsDegeneracy[2][4].values())

        # number of GCs at degenerate third codon positions
        self.mNDGC3 = self.mN2DGC3 + self.mN3DGC3 + self.mN4DGC3

        if self.mLength > 0:
            self.mPNGC = "%5.4f" % (float(self.mNGC) / float(self.mLength))
        else:
            self.mPNGC = "na"

        if self.mNCodons > 0:
            self.mPNGC3 = "%5.4f" % (float(self.mNGC3) / float(self.mNCodons))
        else:
            self.mPNGC3 = "na"

        if self.mNSites2D > 0:
            self.mPN2DGC3 = "%5.4f" % (
                float(self.mN2DGC3) / float(self.mNSites2D))
        else:
            self.mPN2DGC3 = "na"

        if self.mNSites3D > 0:
            self.mPN3DGC3 = "%5.4f" % (
                float(self.mN3DGC3) / float(self.mNSites3D))
        else:
            self.mPN3DGC3 = "na"

        if self.mNSites4D > 0:
            self.mPN4DGC3 = "%5.4f" % (
                float(self.mN4DGC3) / float(self.mNSites4D))
        else:
            self.mPN4DGC3 = "na"

        if self.mNSitesD3 > 0:
            self.mPNDGC3 = "%5.4f" % (
                float(self.mNDGC3) / float(self.mNSitesD3))
        else:
            self.mPNDGC3 = "na"

    def getFields(self):
        fields = SequencePropertiesLength.getFields(self)
        return fields + ["%i" % self.mNStopCodons,
                         "%i" % self.mNSites1D,
                         "%i" % self.mNSites2D,
                         "%i" % self.mNSites3D,
                         "%i" % self.mNSites4D,
                         "%i" % self.mNSitesD3,
                         "%i" % self.mNGC,
                         "%i" % self.mNGC3,
                         "%i" % self.mNDGC3,
                         "%i" % self.mN2DGC3,
                         "%i" % self.mN3DGC3,
                         "%i" % self.mN4DGC3,
                         "%s" % self.mPNGC,
                         "%s" % self.mPNGC3,
                         "%s" % self.mPNDGC3,
                         "%s" % self.mPN2DGC3,
                         "%s" % self.mPN3DGC3,
                         "%s" % self.mPN4DGC3]

    def getHeaders(self):
        headers = SequencePropertiesLength.getHeaders(self)
        return headers + ["nstops",
                          "nsites1d",
                          "nsites2d",
                          "nsites3d",
                          "nsites4d",
                          "nsitesd3",
                          "ngc",
                          "ngc3",
                          "ndgc3",
                          "n2dgc3",
                          "n3dgc3",
                          "n4dgc3",
                          "pgc",
                          "pgc3",
                          "pdgc3",
                          "p2dgc3",
                          "p3dgc3",
                          "p4dgc3",
                          ]


class SequencePropertiesAA(SequenceProperties):
    """Add Properties : amino acid composition of nucleotide sequence.

    The codons in the nucleotide sequence are translated into amino
    acids before counting. The nucleotide sequence must be a multiple
    of 3.

    nA, nC, nD, ...
        Amino acid counts.
    pA, pC, pD, ...
        Amino acid frequencies.

    """

    def __init__(self, reference_usage=[]):

        SequenceProperties.__init__(self)

        self.mReferenceUsage = reference_usage

        # counts of amino acids
        self.mCountsAA = {}
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            self.mCountsAA[x] = 0

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)

        for aa, count in list(other.mCountsAA.items()):
            self.mCountsAA[aa] += count

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""

        SequenceProperties.loadSequence(self, sequence, seqtype)

        if len(sequence) % 3:
            raise ValueError(
                '''sequence length is not a multiple of 3 (length=%i)''' %
                (len(sequence)))

        # counts of amino acids
        self.mCountsAA = {}

        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            self.mCountsAA[x] = 0

        for codon in (sequence[x:x + 3] for x in range(0, len(sequence), 3)):
            aa = Genomics.MapCodon2AA(codon)
            self.mCountsAA[aa] += 1

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        t = 0
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append("%i" % self.mCountsAA[x])
            t += self.mCountsAA[x]
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append("%f" % (float(self.mCountsAA[x]) / t))
        return fields

    def getHeaders(self):
        '''Return list of data headers'''
        headers = SequenceProperties.getHeaders(self)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            headers.append("n%s" % x)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            headers.append("p%s" % x)
        return headers


class SequencePropertiesAminoAcids(SequenceProperties):
    """Add Properties : amino acid composition

    nA, nC, nD, ...
        Amino acid counts.
    pA, pC, pD, ...
        Amino acid frequencies.
    """

    def __init__(self, reference_usage=[]):

        SequenceProperties.__init__(self)

        # counts of amino acids
        self.mCountsAA = {}
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            self.mCountsAA[x] = 0
        self.mOtherCounts = 0

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)

        for aa, count in list(other.mCountsAA.items()):
            self.mCountsAA[aa] += count

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence. """

        SequenceProperties.loadSequence(self, sequence, seqtype)

        # set to zero
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            self.mCountsAA[x] = 0
        self.mOtherCounts = 0

        for aa in sequence:
            if aa == "-":
                continue
            try:
                self.mCountsAA[aa] += 1
            except KeyError:
                self.mOtherCounts += 1

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        t = 0

        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append("%i" % self.mCountsAA[x])
            t += self.mCountsAA[x]

        if t > 0:
            for x in Bio.Alphabet.IUPAC.extended_protein.letters:
                fields.append("%f" % (float(self.mCountsAA[x]) / t))
        else:
            for x in Bio.Alphabet.IUPAC.extended_protein.letters:
                fields.append("0")

        return fields

    def getHeaders(self):

        fields = SequenceProperties.getHeaders(self)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append("n%s" % x)
        for x in Bio.Alphabet.IUPAC.extended_protein.letters:
            fields.append("p%s" % x)

        return fields


class SequencePropertiesCodons(SequencePropertiesLength):
    """Add Properties : codon frequencies

    nAAA, nAAC, ...
         Codon counts
    pAAA, pAAC, ...
         Codon frequencies
    """

    def __init__(self):

        SequencePropertiesLength.__init__(self)

        self.mCodonCounts = {}
        for c in list(Genomics.GeneticCodeAA.keys()):
            self.mCodonCounts[c] = 0

    def addProperties(self, other):
        SequencePropertiesLength.addProperties(self, other)

        for codon, count in list(other.mCodonCounts.items()):
            if codon not in self.mCodonCounts:
                self.mCodonCounts[codon] = 0
            self.mCodonCounts[codon] += count

    def updateProperties(self):

        SequencePropertiesLength.updateProperties(self)

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""

        if len(sequence) % 3:
            raise ValueError(
                '''sequence length is not a multiple of 3
                (length=%i)''' % (len(sequence)))

        SequencePropertiesLength.loadSequence(self, sequence, seqtype)

        # uppercase all letters and count codons
        self.mCodonCounts = Genomics.CountCodons(sequence.upper())

    def getFields(self):

        fields = SequencePropertiesLength.getFields(self)

        keys = list(self.mCodonCounts.keys())
        keys.sort()
        t = 0
        for key in keys:
            t += self.mCodonCounts[key]
            fields.append("%i" % self.mCodonCounts[key])
        t = float(t)
        for key in keys:
            fields.append("%f" % (float(self.mCodonCounts[key]) / t))

        return fields

    def getHeaders(self):

        fields = SequencePropertiesLength.getHeaders(self)

        keys = list(self.mCodonCounts.keys())
        keys.sort()
        for key in keys:
            fields.append("n%s" % key)
        for key in keys:
            fields.append("p%s" % key)

        return fields


class SequencePropertiesCodonUsage(SequencePropertiesCodons):
    """Add properties : Codon usage

    The codon frequency is the ratio of the number of times a
    particular codon is used for a particular amino acid, didived the
    number of times that particular amino acid appears in the
    sequence. A ratio of 1.0 means that this particular codon is
    always used to encode its amino acid, while a frequency of 0.5
    means it is used 50% of the times.

    rAAA, rAAC, ...
        Codon frequencies.

    """

    def __init__(self):

        SequencePropertiesCodons.__init__(self)

        self.mCodonFrequencies = {}

        for codon, aa in list(Genomics.GeneticCodeAA.items()):
            self.mCodonFrequencies[codon] = 0
        for codon in Genomics.StopCodons:
            self.mCodonFrequencies[codon] = 0

    def addProperties(self, other):

        SequencePropertiesCodons.addProperties(self, other)

    def updateProperties(self):

        SequencePropertiesCodons.updateProperties(self)

        self.mCodonFrequencies = Genomics.CalculateCodonFrequenciesFromCounts(
            self.mCodonCounts)

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        keys = list(self.mCodonFrequencies.keys())
        keys.sort()
        for key in keys:
            fields.append("%f" % self.mCodonFrequencies[key])

        return fields

    def getHeaders(self):

        fields = SequenceProperties.getHeaders(self)

        keys = list(self.mCodonFrequencies.keys())
        keys.sort()
        for key in keys:
            fields.append("r%s" % key)

        return fields


class SequencePropertiesCodonTranslator(SequencePropertiesCodonUsage):
    """Add properties : codon sequence is translated into frequencies.

    tsequence
        comma separated list of codon frequencies. The frequencies are
        in percentages.
    """

    def __init__(self):

        SequencePropertiesCodonUsage.__init__(self)

        self.mSequence = ""

    def addProperties(self, other):

        self.mSequence += other.mSequence

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""
        SequencePropertiesCodons.loadSequence(self, sequence, seqtype)

        self.mSequence = sequence

    def getHeaders(self):

        fields = SequenceProperties.getHeaders(self)
        fields.append("tsequence")
        return fields

    def getFields(self):

        fields = SequenceProperties.getFields(self)

        data = []
        for x in range(0, len(self.mSequence), 3):
            codon = self.mSequence[x:x + 3]
            if codon in self.mCodonFrequencies:
                v = self.mCodonFrequencies[codon]
            else:
                v = 0.0

            data.append("%i" % (v * 100))

        fields.append(",".join(data))

        return fields


class SequencePropertiesBias(SequencePropertiesCodons):
    """Add properties : bias measures of codon sequence.

    This class outputs metrics showing how biased the codon usage in a
    particular sequence is compared to a reference codon usage.  The
    reference codon usage is given as a dictionary of codon
    frequencies and multiple dictionaries can be given to compute the
    bias against multiple codon usages.

    entropy
        Entropy of the sequence.
    ml0, ml1, ...
        Message length of sequence compared to reference codon
        usages.
    relml0, relml1, ...
        Relative message length of sequence compared to reference codon
        usages. The relative message length is the message lenght divided
        by the number of codons.
    relentropy0, relentropy1, ...
        Relative entropy of sequence compared to reference codon usages.
        Also called conditional entropy or encoding cost.
    kl0, kl1, ...
        Kullback-Leibler Divergence (relative entropy) of sequence compared
        to reference codon usages.

    Arguments
    ---------

    reference_usage : list
        A list of codon frequency tables. The bias will be computed
        against each.
    pseudocounts : int
        Pseudo-counts to add
    """

    def __init__(self, reference_usage=[], pseudocounts=0):

        SequencePropertiesCodons.__init__(self)
        self.mReferenceUsage = reference_usage
        self.mEntropy = 0
        self.mPseudoCounts = pseudocounts

    def updateProperties(self):

        SequenceProperties.updateProperties(self)
        self.mProperties = []
        self.mEntropy = self.getEntropy()
        for usage in self.mReferenceUsage:
            self.mProperties.append({'ml': self.getMessageLength(usage),
                                     'entropy': self.getEntropy(usage),
                                     'kl': self.getKL(usage)})

    def getMessageLength(self, usage):
        """return message length of a sequence
        in terms of its reference usage."""
        ml = 0
        for codon, count in list(self.mCodonCounts.items()):
            ml -= float(count) * math.log(usage[codon])
        return ml

    def getEntropy(self, usage=None):
        """return entropy of a source in terms of a reference usage.
        Also called conditional entropy or encoding cost.

        Note that here I compute the sum over 20 entropies,
        one for each amino acid.

        If not given, calculate entropy.
        """

        e = 0
        freqs = Genomics.CalculateCodonFrequenciesFromCounts(
            self.mCodonCounts, self.mPseudoCounts)
        if usage is None:
            usage = freqs
        for codon, count in list(self.mCodonCounts.items()):
            e -= freqs[codon] * math.log(usage[codon])
        return e

    def getKL(self, usage):
        """return Kullback-Leibler Divergence (relative entropy) of sequences with
        respect to reference codon usage.
        """
        e = 0
        freqs = Genomics.CalculateCodonFrequenciesFromCounts(
            self.mCodonCounts, self.mPseudoCounts)
        for codon, count in list(self.mCodonCounts.items()):
            e += usage[codon] * math.log(usage[codon] / freqs[codon])
        return e

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        fields.append("%f" % self.mEntropy)
        for x in self.mProperties:
            fields.append("%f" % x['ml'])
            fields.append("%f" % (x['ml'] / self.mNCodons))
            fields.append("%f" % x['entropy'])
            fields.append("%f" % x['kl'])
        return fields

    def getHeaders(self):

        headers = SequenceProperties.getHeaders(self)
        headers.append("entropy")
        for x in range(len(self.mReferenceUsage)):
            headers.append("ml%i" % x)
            headers.append("relml%i" % x)
            headers.append("relentropy%i" % x)
            headers.append("kl%i" % x)
        return headers


class SequencePropertiesCounts(SequenceProperties):
    """Add Properties : Residue counts against arbirtrary alphabet

    nUnk
        Number of unknown residues
    nA, nB, ...
        Character counts
    pA, pB, ...
        Character frequencies

    Arguments
    ---------
    alphabet : string
        List of characters in alphabet

    """

    def __init__(self, alphabet):

        SequenceProperties.__init__(self)
        self.mAlphabet = alphabet
        self.mCounts = {}
        self.mCountsOthers = 0

        for x in self.mAlphabet:
            self.mCounts[x] = 0

    def addProperties(self, other):
        SequenceProperties.addProperties(self, other)
        for na, count in list(other.mCounts.items()):
            self.mCounts[na] += count
        self.mCountsOthers += self.mCountsOthers

    def loadSequence(self, sequence, seqtype="na"):
        """load sequence properties from a sequence."""

        SequenceProperties.loadSequence(self, sequence, seqtype)

        # counts of amino acids
        self.mCounts = {}
        for x in self.mAlphabet:
            self.mCounts[x] = 0
        self.mCountsOthers = 0

        for na in sequence.upper():
            try:
                self.mCounts[na] += 1
            except KeyError:
                self.mCountsOthers += 1

    def getFields(self):
        fields = SequenceProperties.getFields(self)
        fields.append("%i" % self.mCountsOthers)
        t = 0

        for x in self.mAlphabet:
            fields.append("%i" % self.mCounts[x])
            t += self.mCounts[x]

        if t == 0:
            for x in self.mAlphabet:
                fields.append("na")
        else:
            for x in self.mAlphabet:
                fields.append("%f" % (float(self.mCounts[x]) / t))

        return fields

    def getHeaders(self):
        headers = SequenceProperties.getHeaders(self)
        headers.append("nUnk")
        headers.extend(["n%s" % x for x in self.mAlphabet])
        headers.extend(["p%s" % x for x in self.mAlphabet])
        return headers


class SequencePropertiesEntropy(SequencePropertiesCounts):
    """Add properties : Entropy of a sequence

    entropy
        Entropy of the sequence

    Arguments
    ---------
    alphabet : string
        List of characters in alphabet
    pseudocounts : int
        Pseudo-counts to add

    """

    def __init__(self, alphabet, pseudocounts=0):

        SequencePropertiesCounts.__init__(self, alphabet=alphabet)
        self.mPseudoCounts = pseudocounts
        self.mEntropy = None

    def addProperties(self, other):
        SequencePropertiesCounts.addProperties(self, other)

    def updateProperties(self):

        SequencePropertiesCounts.updateProperties(self)

        self.mProperties = []

        self.mEntropy = self.getEntropy()

    def getEntropy(self, usage=None):
        """return entropy of a source in terms of a reference usage.

        Also called conditional entropy or encoding cost.
        """

        e = None

        v = list(self.mCounts.values())
        total = sum(v) + len(v) * self.mPseudoCounts
        pc = self.mPseudoCounts
        if total != 0:
            e = 0
            frequencies = [float(x + pc) / total for x in v]
            for f in frequencies:
                if f > 0:
                    e -= f * math.log(f)
        return e

    def getFields(self):

        fields = SequenceProperties.getFields(self)
        if self.mEntropy is not None:
            fields.append("%f" % self.mEntropy)
        else:
            fields.append("na")

        return fields

    def getHeaders(self):

        fields = SequenceProperties.getHeaders(self)
        fields.append("entropy")
        return fields



'''gtf2table.py - annotate genes/transcripts
============================================

:Tags: Genomics Genesets GTF Annotation

Purpose
-------

annotate genes or transcripts given in a :term:`gtf` formatted file
and output them in tabular format.

Annotations can be either computed per gene (all exons across all
transcripts) or per transcript. The input needs to be sorted
accordingly.

The script iterates over each gene or transcript model in turn and
outputs a row in a table for each. Multiple annotations can be
computed at the same time adding additional columns to the table.

For example, to output information about exons in genes::

    zcat in.gtf.gz | cgat gtf2table -v 0 --counter=length

+---------------+----+----+----+---------+------+--------+----+----+----+
|gene_id        |nval|min |max |mean     |median|stddev  |sum |q1  |q3  |
+---------------+----+----+----+---------+------+--------+----+----+----+
|ENSG00000225373|4   |39  |688 |253.0000 |142.5 |261.5922|1012|58  |688 |
+---------------+----+----+----+---------+------+--------+----+----+----+
|ENSG00000267111|1   |744 |744 |744.0000 |744.0 |0.0000  |744 |744 |744 |
+---------------+----+----+----+---------+------+--------+----+----+----+
|ENSG00000267588|3   |135 |489 |370.0000 |486.0 |166.1746|1110|135 |489 |
+---------------+----+----+----+---------+------+--------+----+----+----+
|ENSG00000220978|1   |138 |138 |138.0000 |138.0 |0.0000  |138 |138 |138 |
+---------------+----+----+----+---------+------+--------+----+----+----+

The table contains exon length statistics of each gene, the number (nval),
min, max, and total length(sum) of exons per gene. To count per transcript::

    zcat in.gtf.gz | cgat gtf2table -v 0 --reporter=transcripts --counter=length

+---------------+----+---+---+--------+------+--------+----+---+---+
|transcript_id  |nval|min|max|mean    |median|stddev  |sum |q1 |q3 |
+---------------+----+---+---+--------+------+--------+----+---+---+
|ENST00000592209|3   |58 |227|149.6667|164.0 |69.7344 |449 |58 |227|
+---------------+----+---+---+--------+------+--------+----+---+---+
|ENST00000589741|2   |71 |312|191.5000|191.5 |120.5000|383 |71 |312|
+---------------+----+---+---+--------+------+--------+----+---+---+
|ENST00000391654|2   |39 |583|311.0000|311.0 |272.0000|622 |39 |583|
+---------------+----+---+---+--------+------+--------+----+---+---+
|ENST00000587045|1   |173|173|173.0000|173.0 |0.0000  |173 |173|173|
+---------------+----+---+---+--------+------+--------+----+---+---+

To add also information about cpg-composition, add another counter::

    zcat in.gtf.gz | cgat gtf2table -v 0 --genome=hg19 --reporter=transcripts --counter=length --counter=composition-cpg

+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+
|transcript_id  |nval|min|max|mean    |median|stddev  |sum |q1 |q3 |CpG_count |CpG_density |CpG_ObsExp |
+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+
|ENST00000592209|3   |58 |227|149.6667|164.0 |69.7344 |449 |58 |227|4         |0.01781     |0.13946    |
+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+
|ENST00000589741|2   |71 |312|191.5000|191.5 |120.5000|383 |71 |312|5         |0.02610     |0.17067    |
+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+
|ENST00000391654|2   |39 |583|311.0000|311.0 |272.0000|622 |39 |583|4         |0.01286     |0.09402    |
+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+
|ENST00000587045|1   |173|173|173.0000|173.0 |0.0000  |173 |173|173|1         |0.01156     |0.11396    |
+---------------+----+---+---+--------+------+--------+----+---+---+----------+------------+-----------+

Note that we had to use the ``--genome-file`` option to supply the
genomic sequence.

Additional switches permit counting introns instead of exons (option
``--section``).

Annotations
-----------

The annotations might be derived from a property of the transcript or
gene model itself (such as length, number of exons, etc), the genomic
sequence (such as composition, etc) or require additional data sets.
For example, to compute the coverage with reads, a :term:`bam` formatted
file is required (see option ``--bam-file``). Or to compute the overlap
with genomic densities, a :term:`bigwig` formatted file is required (see
option ``--bigwig-file``).

This section lists the counters available. They are grouped by data sources
they require.

Generic annotations
+++++++++++++++++++

Generic annotations require no additional data source to annotate a
transcript or gene.

length
   output exon length summary of transcript/gene. This counter outputs the
   number of exons in each transcript/gene together with exon length summary
   statistics (minimal exon length, maximal exon length, total exon length).

position
   output genomic coordinates of transcript/gene (chromosome, start, end).

Sequence derived annotations
++++++++++++++++++++++++++++

Sequence derived annotations require the genomic sequence to compute
properties of the gene/transcript model (see option ``--genome-file``).

composition-na
   output nucleotide composition of transcript/gene

composition-cgp
   output CpG count, CpG density and CpG observed / expected for
   each transcript/gene.

splice
   output splicing summary of transcript/gene. Outputs the number of
   canonical and non-canonical splice sites.

quality
   output base-quality information summary of gene. Needs quality scores.

Interval derived annotations
++++++++++++++++++++++++++++

Annotations in the list below relate a gene or transcript to a set of
intervals given as a second file. The second set of intervals is given
by the option ``filename-gff``. By default, the intervals are expected
to be given as a :term:`gff` formatted file, but alternative formats
(:term:`gtf` and :term:`bed`) are possible (see option
``--filename-format``)

overlap
    compute overlap of genes in input with features in second stream.
    Requires a :term:`gff` formatted file with gene territories.

overlap-stranded
    count overlap with genomic features in second another file. Outputs
    the number of overlapping exons. Records the direction of overlap
    (sense/antisense). Requires a :term:`gff` formatted file with
    features.

territories
    compute overlap of transcripts/genes with territories. Territories
    are genomic, non-overlapping intervals. For each transcript the
    counter outputs the status (U=unique match, A=ambiguous match
    (overlap with multiple territories, 0=no match), the number of
    territories it overlaps and the territories it overlaps.

coverage

   compute the coverage - per nucleotide - of the gene/transcript
   models with intervals given in ``--gff-file``.  Coverage values
   are output in 5' to 3' together with summary statistics (bases
   covered, minimum, maximum coverage, etc.). By using the options
   ``--restrict-feature`` or ``--restrict-source`` the counting can be
   rescricted to particular features or sources in the :term:`gff` file.

distance

   compute distance of genes to features in a second file. Requires a
   second :term:`gff` formatted file. The strand information of the
   features is ignored.

binding-pattern

   given a list of intervals, determine the binding pattern within and
   surrounding the gene. For each gene, intervals overlapping the CDS,
   introns, UTRs and the flank are collected and recorded. The binding
   is summarized with a binding pattern, a binary pattern indicating
   overlap/no overlap with 5' flank, 5' UTR, CDS, Introns, 3' UTR, 3'
   flank. This method is useful to check where transcription factor binding
   sites are located aronud a gene/transcript model.

proximity

   report summary stats (lengths, values) of features in proximity to
   genes input gene set. Requires a :term:`gff` formatted file with
   genomic features. This feature is useful when aiming to normalize a
   value, such as a substitution rate of a transcript model, by
   substitution rates of segments in the neighbourhood such as
   ancestral repeats. The values are given in the ``score`` field of
   the :term:`gff` formatted file. The radius for proximity is controlled
   by the option ``--proximal-distance``.

proximity-exclusive

   as proximity, but exclude any ranges in the :term:`gff` formatted file that
   overlap the transcript/gene model.

proximity-lengthmatched

   as proximity-exclusive, but length-match features with
   genes. Segments are declared equal in length if they are within
   10% of the original segments length.

neighbours
    output features in second stream that are in proximity to genes
    in input. This is similar to the ``proximity`` counters, but also
    outputs the features that are in proximity.

Gene set derived annotations
++++++++++++++++++++++++++++

distance-genes
   compute distance of genes to genes in a second file. Requires a
   second :term:`gtf` formatted file with genes. The counter distinguishes
   a variety of cases (closest upstream/downstream).

distance-tss
   compute distance of genes to transcription start sites. Requires a
   second :term:`gtf` formatted file with genes.

overlap-transcripts
    count overlap of genes with transcripts in another set.
    Requires a :term:`gtf` formatted file.

overrun
   output intron overrun, exons in the input gene set extending
   into the introns of a reference gene set. Requries a :term:`gtf`
   formatted file with a reference gene set.

splice-comparison
   Compare how splice site usage compares between a gene/transcript
   with transcripts in a reference gene set. Outputs found, missed,
   perfect, partial, incomplete splice sites and exon-skipping events.


Short-read derived annotations
++++++++++++++++++++++++++++++

Short-read derived annotations count the overlap of reads given
in a :term:`bam` formatted file (see option ``--bam-file``) with
gene or transcript models.

read-overlap

   output number of reads overlapping a transcript/gene model.
   Outputs the number of reads overlapping the transcript/gene model.
   Reads are counted separately for sense and antisense overlap.

read-coverage

   output read coverage summary statistics of transcript/gene model.
   Outputs the number of reads overlapping the transcript/gene model,
   the number of bases overlapped by reads and summary statistics of
   the minimimum, maximum, etc. coverage per pase. Reads are counted
   separately for sense and antisense overlap. The counter does not
   take into account splice sites. As it does per-base counting,
   it is slower than ``read-overlap`` and there is no need to
   use both at the same time.

read-extension

   Counter of special interest. This counter outputs the read density
   in bins upstream, within and downstream of transcript models. The
   counter can be used to predict the length of the 3' and 5' UTR.

read-fullcounts

   count number of reads overlapping a gene or transcript. Reads
   are classified according to the type of overlap (sense/antisense),
   full or partial match to exon, splice status (spliced/unspliced).
   See below for more information.

   Unique and non-unique matches are counted (by alignment start
   position).

read-counts

   count number of reads overlapping a gene or transcript. Summarizes
   the output of read-fullcounts.

readpair-fullcounts

   count number of read pairs overlapping a gene or transcript. Pairs
   are counted according to a variety of attributes (exonic
   overlap/read pair status/splice status/...). See below for more
   information.

readpair-counts

   count number of read pairs overlapping a gene or
   transcript. Summarizes the output of readpair-fullcounts.


Classifiers
-----------

Classifiers not only annotate the transcripts or gene model, but also
aim to provide some classification based on these annotations. They
require a secondary file (see option ``--gff-file``) for the
classification.

classifier

   classify transcripts according to genomic annotation.  Requires a
   :term:`gff` file with genomic annotations (see :doc:`gtf2gff`)
   delineating genomic regions as intronic, intergenic, exonic,
   etc. This is useful for a rough classification of transcripts as
   intergenic, intronic, etc.

   Best used for ChIP-Seq data sets (each peak is a "transcript"). The
   method is a little out of place in this script, but is here as it
   uses much of the code implemented here.

classifier-rnaseq

   classify transcripts with respect to a reference geneset. The
   classifiers aims to match a transcript up with the "most similar"
   transcript in the reference gene set and dependending on
   attributes, classifies it as a good match, alternative transcript,
   fragment, etc. The classifier takes into account strand and prefers
   sense matches to antisense matches.

classifier-polii

   classify according to PolII transcripts. A gene/transcript is
   transcribed, if it is covered by large PolII intervals over 80% of
   its length. A gene/transript is primed if its promotor/UTR is
   covered by 50% of its length, while the rest of the gene body
   isn't.

Other annotations
++++++++++++++++++

The following methods (see option ``--counter``) are available:

bigwig-counts

   collect density values from a :term:`bigwig` formatted file and output
   summary statistics and percentage of bases covered (``pcovered``)
   by value in bigwig file. Requires option ``--bigwig-file``.


Read counting
-------------

The methods ``read-counts`` and ``readpair-counts`` count the number
of reads inside a :term:`bam` formatted file that map inside a
transcript or gene.

These counters proceed on a per-gene or per-transcript basis depending
on the ``--reporter`` option. All reads overlapping the exons or
introns of a transcript are collected and counted either individually
(``read-counts``) or as pairs (``readpair-counts``).

For paired-read counting, each pair is classified and counted
according to the following for axes:

1. Pair status: proper pair, unmapped read in pair, ...
2. Direction: for stranded protocols, sense, antisense, ...
3. Overlap status: Pair overlaps exons only, introns only, ...
4. Splice status: Reads are unspliced, spliced consistently with
   transcript model, or inconsistently with transcript model

Counts are then collated into a smaller set of summaries for convenience:

+--------------------+----------------------------------------------------+
|Column              |Number of read pairs                                |
+--------------------+----------------------------------------------------+
|counted_all         |considered to be correct: they are sense, overlap   |
|                    |with exons fully.                                   |
+--------------------+----------------------------------------------------+
|counted_spliced     |in sense direction and exonic, spliced and spliced  |
|                    |correctly.                                          |
|                    |                                                    |
+--------------------+----------------------------------------------------+
|counted_unspliced   |in sense direction and exonic, not spliced.         |
+--------------------+----------------------------------------------------+
|sense_intronic      |intronic and sense direction.                       |
+--------------------+----------------------------------------------------+
|sense_inconsistent  |in sense direction, but overlap both introns and    |
|                    |exons or extend beyond the transcript modelse       |
+--------------------+----------------------------------------------------+
|sense_other         |in sense direction, but not counted, intronic or    |
|                    |inconsistent.                                       |
+--------------------+----------------------------------------------------+
|antisense           |in antisense direction.                             |
+--------------------+----------------------------------------------------+
|nonsense            |in unexpected orientation                           |
+--------------------+----------------------------------------------------+
|notproper           |that are not proper pairs                           |
+--------------------+----------------------------------------------------+
|quality_pairs       |affected by a the removal of a low-quality read     |
+--------------------+----------------------------------------------------+
|quality_reads       |Number of low quality reads removed                 |
+--------------------+----------------------------------------------------+
|total               |Total number of pairs considered                    |
+--------------------+----------------------------------------------------+

The Counter ``readpair-fullcounts`` provides the detailed counts
for each transcript model according to the four axes. The output
can be used to implement custom counting schemes.

Reads below a minimum quality score will be ignored
(``--min-read-quality``). By default, all reads will be counted.

Reads mapping to multiple locations will be downweighted if
the ``--weight-multi-mapping`` option is set. This requires
the presence of the ``NH`` flag in the :term:`bam` file.

For paired read counting, the library type can be specified with the
``--library-type`` option to make use of strand information. Library
types are labelled according to the tophat_ and cufflinks_
convention. A summary is `here
<http://www.nature.com/nprot/journal/v7/n3/fig_tab/nprot.2012.016_T1.html>`_.

+--------------------+--------------------+------------------------------+
|Library type        |RNA-seq protocol    |Explanation                   |
+--------------------+--------------------+------------------------------+
|fr-unstranded       |Illumina TruSeq     |Reads from the leftmost end of|
| (default)          |                    |the fragment (in transcript   |
|                    |                    |coordinates) map to the       |
|                    |                    |transcript strand, and the    |
|                    |                    |rightmost end maps to the     |
|                    |                    |opposite strand left          |
|                    |                    |                              |
+--------------------+--------------------+------------------------------+
|fr-firststrand      |dUTP, NSR, NNSR39   |Same as above except we       |
|                    |                    |enforce the rule that the     |
|                    |                    |rightmost end of the fragment |
|                    |                    |(in transcript coordinates) is|
|                    |                    |the first sequenced (or only  |
|                    |                    |sequenced for single-end      |
|                    |                    |reads). Equivalently, it is   |
|                    |                    |assumed that only the strand  |
|                    |                    |generated during first strand |
|                    |                    |synthesis is sequenced        |
+--------------------+--------------------+------------------------------+
|fr-secondstrand     |Directional Illumina|Same as above                 |
|                    |(Ligation)          |except TopHat/Cufflinks       |
|                    |Standard SOLiD      |enforce the rule that the     |
|                    |                    |leftmost end of the fragment  |
|                    |                    |(in transcript coordinates) is|
|                    |                    |the first sequenced (or only  |
|                    |                    |sequenced for single-end      |
|                    |                    |reads). Equivalently, it is   |
|                    |                    |assumed that only the strand  |
|                    |                    |generated during second strand|
|                    |                    |synthesis is sequenced        |
+--------------------+--------------------+------------------------------+

For unstranded protocols, all reads and pairs are considered to matching
in the sense direction.

Usage
-----

Type::

   python gtf2table.py --help

for command line help.

Command line options
--------------------

'''

import sys
import pysam

import CGAT.Experiment as E
import CGAT.GTF as GTF
import CGAT.IndexedFasta as IndexedFasta
import CGAT.GeneModelAnalysis as GeneModelAnalysis

import pyBigWig


def main(argv=None):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default].")

    parser.add_option("-q", "--quality-file",
                      dest="quality_file",
                      type="string",
                      help="filename with genomic base quality "
                      "information [default=%default].")

    parser.add_option("-b", "--bam-file", dest="bam_files",
                      type="string", metavar="bam",
                      help="filename with read mapping information. "
                      "Multiple files can be submitted in a "
                      "comma-separated list [default=%default].")

    parser.add_option("-i", "--bigwig-file", dest="bigwig_file",
                      type="string", metavar="bigwig",
                      help="filename with bigwig information "
                      "[default=%default].")

    parser.add_option("-f", "--gff-file", dest="filename_gff",
                      type="string", action="append", metavar='bed',
                      help="filename with extra gff files. The order "
                      "is important [default=%default].")

    parser.add_option("--filename-format", dest="filename_format",
                      type="choice",
                      choices=("bed", "gff", "gtf"),
                      help="format of secondary stream [default=%default].")

    parser.add_option("--restrict-source", dest="gff_sources", type="string",
                      action="append",
                      help="restrict input to this 'source' in extra "
                      "gff file (for counter: overlap) [default=%default].")

    parser.add_option("--restrict-feature", dest="gff_features", type="string",
                      action="append",
                      help="restrict input to this 'feature' in extra gff "
                      "file (for counter: overlap) [default=%default].")

    parser.add_option("-r", "--reporter", dest="reporter", type="choice",
                      choices=("genes", "transcripts"),
                      help="report results for 'genes' or 'transcripts' "
                      "[default=%default].")

    parser.add_option("-s", "--section", dest="sections",
                      type="choice",
                      action="append",
                      choices=("exons", "introns"),
                      help="select range on which counters will operate "
                      "[default=%default].")

    parser.add_option("-c", "--counter", dest="counters",
                      type="choice",
                      action="append",
                      choices=(	"bigwig-counts",
                                "binding-pattern",
                                "classifier",
                                "classifier-rnaseq",
                                "classifier-rnaseq-splicing",
                                "classifier-polii",
                                "composition-na",
                                "composition-cpg",
                                "coverage",
                                "distance",
                                "distance-genes",
                                "distance-tss",
                                "length",
                                'neighbours',
                                "overlap",
                                "overlap-stranded",
                                "overlap-transcripts",
                                "overrun",
                                "position",
                                "proximity",
                                "proximity-exclusive",
                                "proximity-lengthmatched",
                                "quality",
                                "read-coverage",
                                "read-extension",
                                "read-overlap",
                                "read-counts",
                                "read-fullcounts",
                                "readpair-counts",
                                "readpair-fullcounts",
                                "splice",
                                "splice-comparison",
                                "territories"),
                      help="select counters to apply to input "
                      "[default=%default].")

    parser.add_option("--add-gtf-source", dest="add_gtf_source",
                      action="store_true",
                      help="add gtf field of source to output "
                      "[default=%default].")

    parser.add_option("--proximal-distance", dest="proximal_distance",
                      type="int",
                      help="distance to be considered proximal to "
                      "an interval [default=%default].")

    parser.add_option("--multi-mapping-method",
                      dest="multi_mapping",
                      type="choice",
                      choices=('all', 'ignore', 'weight'),
                      help="how to treat multi-mapping reads in "
                      "bam-files. Requires "
                      "the NH flag to be set by the mapper "
                      "[default=%default].")

    parser.add_option("--use-barcodes",
                      dest="use_barcodes",
                      action="store_true",
                      help="Use barcodes to count unique umi's. "
                      "UMI's are specified in the read identifier "
                      "as the last field, where fields are separated "
                      "by underscores, e.g. "
                      "@READ:ILLUMINA:STUFF_NAMINGSTUFF_UMI. "
                      "When true, unique counts are returned. "
                      "Currently only compatible with count-reads")

    parser.add_option("--sample-probability",
                      dest="sample_probability",
                      type="float",
                      help="Specify the probability of whether any"
                      "given read or read pair in a file bam is counted"
                      "Currently only compatible with count-reads")

    parser.add_option("--column-prefix", dest="prefixes",
                      type="string",
                      action="append",
                      help="add prefix to column headers - prefixes "
                      "are used in the same order as the counters "
                      "[default=%default].")

    parser.add_option("--library-type",
                      dest="library_type",
                      type="choice",
                      choices=("unstranded",
                               "firststrand",
                               "secondstrand",
                               "fr-unstranded",
                               "fr-firststrand",
                               "fr-secondstrand"),
                      help="library type of reads in bam file. "
                      "[default=%default]")

    parser.add_option("--min-mapping-quality",
                      dest="minimum_mapping_quality",
                      type="float",
                      help="minimum mapping quality. Reads with a quality "
                      "score of less will be ignored. "
                      "[default=%default]")

    parser.set_defaults(
        genome_file=None,
        reporter="genes",
        with_values=True,
        sections=[],
        counters=[],
        filename_gff=[],
        filename_format=None,
        gff_features=[],
        gff_sources=[],
        add_gtf_source=False,
        proximal_distance=10000,
        bam_files=None,
        multi_mapping='all',
        library_type='fr-unstranded',
        prefixes=[],
        minimum_mapping_quality=0,
        use_barcodes=False,
        sample_probability=1.0
    )

    if not argv:
        argv = sys.argv

    (options, args) = E.Start(parser, add_output_options=True, argv=argv)

    if options.prefixes:
        if len(options.prefixes) != len(options.counters):
            raise ValueError(
                "if any prefix is given, the number of prefixes "
                "must be the same as the number of counters")

    # get files
    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta(options.genome_file)
    else:
        fasta = None

    if options.quality_file:
        quality = IndexedFasta.IndexedFasta(options.quality_file)
        quality.setTranslator(IndexedFasta.TranslatorBytes())
    else:
        quality = None

    if options.bam_files:
        bam_files = []
        for bamfile in options.bam_files.split(","):
            bam_files.append(pysam.AlignmentFile(bamfile, "rb"))
    else:
        bam_files = None

    if options.bigwig_file:
        bigwig_file = pyBigWig.open(options.bigwig_file)
    else:
        bigwig_file = None

    counters = []

    if not options.sections:
        E.info("counters will use the default section (exons)")
        options.sections.append(None)

    if not options.gff_sources:
        options.gff_sources.append(None)
    if not options.gff_features:
        options.gff_features.append(None)

    cc = E.Counter()

    for n, c in enumerate(options.counters):
        if options.prefixes:
            prefix = options.prefixes[n]
        else:
            prefix = None

        if c == "position":
            for section in options.sections:
                counters.append(
                    GeneModelAnalysis.CounterPosition(
                        section=section,
                        options=options,
                        prefix=prefix))
        elif c == "length":
            for section in options.sections:
                counters.append(
                    GeneModelAnalysis.CounterLengths(
                        section=section,
                        options=options,
                        prefix=prefix))
        elif c == "splice":
            if fasta is None:
                raise ValueError('splice requires a genomic sequence')
            counters.append(GeneModelAnalysis.CounterSpliceSites(fasta=fasta, prefix=prefix))
        elif c == "quality":
            if fasta is None:
                raise ValueError('quality requires a quality score sequence')
            counters.append(GeneModelAnalysis.CounterQuality(fasta=quality, prefix=prefix))
        elif c == "overrun":
            counters.append(GeneModelAnalysis.CounterOverrun(
                filename_gff=options.filename_gff,
                options=options,
                prefix=prefix))
        elif c == "read-coverage":
            counters.append(GeneModelAnalysis.CounterReadCoverage(
                bam_files,
                options=options,
                prefix=prefix))
        elif c == "read-extension":
            counters.append(GeneModelAnalysis.CounterReadExtension(
                bam_files,
                filename_gff=options.filename_gff,
                options=options,
                prefix=prefix))
        elif c == "read-overlap":
            counters.append(GeneModelAnalysis.CounterReadOverlap(
                bam_files,
                multi_mapping=options.multi_mapping,
                minimum_mapping_quality=options.minimum_mapping_quality,
                options=options,
                prefix=prefix))
        elif c == "read-counts":
            counters.append(GeneModelAnalysis.CounterReadCounts(
                bam_files,
                multi_mapping=options.multi_mapping,
                use_barcodes=options.use_barcodes,
                sample_probability=options.sample_probability,
                minimum_mapping_quality=options.minimum_mapping_quality,
                options=options,
                prefix=prefix))
        elif c == "read-fullcounts":
            counters.append(GeneModelAnalysis.CounterReadCountsFull(
                bam_files,
                multi_mapping=options.multi_mapping,
                sample_probability=options.sample_probability,
                minimum_mapping_quality=options.minimum_mapping_quality,
                options=options,
                prefix=prefix))
        elif c == "readpair-counts":
            counters.append(GeneModelAnalysis.CounterReadPairCounts(
                bam_files,
                multi_mapping=options.multi_mapping,
                sample_probability=options.sample_probability,
                library_type=options.library_type,
                minimum_mapping_quality=options.minimum_mapping_quality,
                options=options,
                prefix=prefix))
        elif c == "readpair-fullcounts":
            counters.append(GeneModelAnalysis.CounterReadPairCountsFull(
                bam_files,
                multi_mapping=options.multi_mapping,
                sample_probability=options.sample_probability,
                minimum_mapping_quality=options.minimum_mapping_quality,
                options=options,
                prefix=prefix))
        elif c == "bigwig-counts":
            counters.append(GeneModelAnalysis.CounterBigwigCounts(
                bigwig_file,
                options=options, prefix=prefix))
        elif c == "splice-comparison":
            if fasta is None:
                raise ValueError('splice-comparison requires a genomic '
                                 'sequence')
            counters.append(GeneModelAnalysis.CounterSpliceSiteComparison(
                fasta=fasta,
                filename_gff=options.filename_gff,
                feature=None,
                source=None,
                options=options, prefix=prefix))
        elif c == "composition-na":
            if fasta is None:
                raise ValueError('composition-na requires a genomic sequence')
            for section in options.sections:
                counters.append(GeneModelAnalysis.CounterCompositionNucleotides(
                    fasta=fasta,
                    section=section,
                    options=options,
                    prefix=prefix))
        elif c == "composition-cpg":
            if fasta is None:
                raise ValueError('composition-cpg requires a genomic sequence')
            for section in options.sections:
                counters.append(GeneModelAnalysis.CounterCompositionCpG(
                    fasta=fasta,
                    section=section,
                    options=options, prefix=prefix))

        elif c in ("overlap",
                   "overlap-stranded",
                   "overlap-transcripts",
                   "proximity",
                   "proximity-exclusive",
                   "proximity-lengthmatched",
                   "neighbours",
                   "territories",
                   "distance",
                   "distance-genes",
                   "distance-tss",
                   "binding-pattern",
                   "coverage"):
            if c == "overlap":
                template = GeneModelAnalysis.CounterOverlap
            if c == "overlap-stranded":
                template = GeneModelAnalysis.CounterOverlapStranded
            elif c == "overlap-transcripts":
                template = GeneModelAnalysis.CounterOverlapTranscripts
            elif c == "proximity":
                template = GeneModelAnalysis.CounterProximity
            elif c == "neighbours":
                template = GeneModelAnalysis.CounterNeighbours
            elif c == "proximity-exclusive":
                template = GeneModelAnalysis.CounterProximityExclusive
            elif c == "proximity-lengthmatched":
                template = GeneModelAnalysis.CounterProximityLengthMatched
            elif c == "territories":
                template = GeneModelAnalysis.CounterTerritories
            elif c == "distance":
                template = GeneModelAnalysis.CounterDistance
            elif c == "distance-genes":
                template = GeneModelAnalysis.CounterDistanceGenes
            elif c == "distance-tss":
                template = GeneModelAnalysis.CounterDistanceTranscriptionStartSites
            elif c == "coverage":
                template = GeneModelAnalysis.CounterCoverage
            elif c == "binding-pattern":
                template = GeneModelAnalysis.CounterBindingPattern

            for section in options.sections:
                for source in options.gff_sources:
                    for feature in options.gff_features:
                        counters.append(template(
                            filename_gff=options.filename_gff,
                            feature=feature,
                            source=source,
                            fasta=fasta,
                            section=section,
                            options=options,
                            prefix=prefix))

        elif c == "classifier":
            counters.append(GeneModelAnalysis.Classifier(
                filename_gff=options.filename_gff,
                fasta=fasta,
                options=options, prefix=prefix))

        elif c == "classifier-rnaseq":
            counters.append(GeneModelAnalysis.ClassifierRNASeq(
                filename_gff=options.filename_gff,
                fasta=fasta,
                options=options, prefix=prefix))
        elif c == "classifier-rnaseq-splicing":
            counters.append(GeneModelAnalysis.ClassifierRNASeqSplicing(
                filename_gff=options.filename_gff,
                fasta=fasta,
                options=options,
                prefix=prefix))
        elif c == "classifier-polii":
            counters.append(GeneModelAnalysis.ClassifierPolII(
                filename_gff=options.filename_gff,
                feature=None,
                source=None,
                fasta=fasta,
                options=options,
                prefix=prefix))
        elif c == "binding-pattern":
            counters.append(GeneModelAnalysis.CounterBindingPattern(
                filename_gff=options.filename_gff,
                feature=None,
                source=None,
                fasta=fasta,
                options=options,
                prefix=prefix))

    if options.reporter == "genes":
        iterator = GTF.flat_gene_iterator
        header = ["gene_id"]
        fheader = lambda x: [x[0].gene_id]
    elif options.reporter == "transcripts":
        iterator = GTF.transcript_iterator
        header = ["transcript_id"]
        fheader = lambda x: [x[0].transcript_id]

    if options.add_gtf_source:
        header.append("source")
        ffields = lambda x: [x[0].source]
    else:
        ffields = lambda x: []

    options.stdout.write("\t".join(
        header + [x.getHeader() for x in counters]) + "\n")

    for gffs in iterator(GTF.iterator(options.stdin)):
        cc.input += 1

        for counter in counters:
            counter.update(gffs)

        skip = len([x for x in counters if x.skip]) == len(counters)
        if skip:
            cc.skipped += 1
            continue

        options.stdout.write("\t".join(
            fheader(gffs) +
            ffields(gffs) +
            [str(counter) for counter in counters]) + "\n")

        cc.output += 1

    E.info("%s" % str(cc))
    for counter in counters:
        E.info("%s\t%s" % (repr(counter), str(counter.counter)))
    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



'''vcf2vcf.py - manipulate vcf files
=================================

Purpose
-------

Manipulate vcf-formatted files.


Usage
-----

Example::

   cat in.vcf | python vcf2vcf.py - --reorder=alphabetical > sorted.vcf

This command generates a sorted vcf with the sample columns in in.vcf
in alphabetical order.

Type::

   python vcf2vcf.py --help

for command line usage.

Methods
-------

This script provides the following methods:

re-order
   reorder sample columns in vcf formatted file according to a given sort order

Documentation
-------------

This is a tool for manipulating vcf-formatted files.  The following
options are available:

+-----------+-------------------------+
|--reorder  |reorders sample columns  |
+-----------+-------------------------+

Reorder
^^^^^^^

To sort sample columns into alphabetical order::

   cat example.vcf | python vcf2vcf.py - --reorder=alphabetical

This will sort the columns in the example.vcf into the order "#CHROM,
POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT, SAMPLE_A, SAMPLE_B,
SAMPLE_C, ..."

To specify a non-alphabetical order::

   cat example.vcf | python vcf2vcf.py - --reorder=SAMPLE_C,SAMPLE_A,SAMPLE_B,...

This will sort the columns in the example.vcf into the order "#CHROM,
POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT, | SAMPLE_C, SAMPLE_A,
SAMPLE_B, ..."

Command line options
--------------------

'''

import sys
import CGAT.Experiment as E
import CGAT.VCF as VCF


def main(argv=sys.argv):

    parser = E.OptionParser(version="%prog version: $Id$",
                            usage=globals()["__doc__"])

    parser.add_option(
        "--method", dest="methods", type="choice", action="append",
        choices=("reorder-samples",),
        help="method to apply [default=%default]")

    parser.add_option(
        "--sort-order", dest="sort_order",
        help="sort order for sample names. Give column names as "
        "comma-separated list or specify ``alphabetical`` "
        "[default=%default]")

    parser.set_defaults(
        methods=[],
        sort_order="alphabetical",
    )

    (options, args) = E.Start(parser, add_pipe_options=True)

    if not options.methods:
        raise ValueError("no method specified")

    infile = VCF.VCFFile(options.stdin)

    sort_order = False
    if "reorder-samples" in options.methods:
        if options.sort_order:
            sort_order = options.sort_order.split(",")
            if "alphabetical" in sort_order:
                sort_order = sorted(infile.samples)

    infile.writeHeader(options.stdout, order=sort_order)

    for vcf in infile:
        if sort_order:
            vcf.order = sort_order
        options.stdout.write(str(vcf) + "\n")

    E.Stop()

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
RateEstimation.py - utilities for computing rate estimates for codon models.
=============================================================================

:Tags: Python

"""

import Bio.Data.CodonTable
try:
    from XGram.Generator.Prebuilt import Codons
    import XGram.Exceptions
except ImportError:
    pass


def evaluateCodonPair(codon1, codon2):
    """evaluate differences between codon pair."""

    codon_table = Bio.Data.CodonTable.standard_dna_table

    changes = []
    for x in range(0, 3):
        if codon1[x] != codon2[x]:
            changes.append((x, codon1[x], codon2[x]))

    if len(changes) != 1:
        return False, False, False

    change = (changes[0][1], changes[0][2])

    is_synonymous = codon_table.forward_table[
        codon1] == codon_table.forward_table[codon2]
    is_transition = change in (("A", "G"), ("G", "A"), ("T", "C"), ("C", "T"))

    return True, is_synonymous, is_transition


def countSubstitutions(pi, Q):
    """count substitituions given a matrix Q and frequencies pi."""

    rI, rV, rS, rN = 0, 0, 0, 0

    codon_table = Bio.Data.CodonTable.standard_dna_table
    codons = list(codon_table.forward_table.keys())

    for codon_i in codons:
        for codon_j in codons:
            if codon_i == codon_j:
                continue

            is_single, is_synonymous, is_transition = evaluateCodonPair(
                codon_i, codon_j)

            v = Q[codon_i][codon_j] * pi[codon_i]

            if is_transition:
                rI += v
            else:
                rV += v

            if is_synonymous:
                rS += v
            else:
                rN += v

    return rI, rV, rS, rN


def initializeQMatrix(codons):
    """get an initialized Q matrix."""

    Q = {}
    for codon_i in codons:
        Q[codon_i] = {}
        for codon_j in codons:
            Q[codon_i][codon_j] = 0.0

    return Q


def getQMatrix(pi, Rsi, Rsv, Rni, Rnv):
    """build a q matrix.

    Diagonal elements are set to the negative of the row sums.
    The matrix is normalized such that trace of the matrix is -1.
    """

    codons = list(Bio.Data.CodonTable.standard_dna_table.forward_table.keys())

    Q = initializeQMatrix(codons)

    trace = 0.0
    for codon_i in codons:
        row_sum = 0.0
        for codon_j in codons:
            if codon_i == codon_j:
                continue

            is_single, is_synonymous, is_transition = evaluateCodonPair(
                codon_i, codon_j)

            if not is_single:
                continue

            if is_synonymous:
                if is_transition:
                    v = Rsi
                else:
                    v = Rsv
            else:
                if is_transition:
                    v = Rni
                else:
                    v = Rnv

            v *= pi[codon_j]
            Q[codon_i][codon_j] = v
            row_sum += v

        Q[codon_i][codon_i] = -row_sum
        trace += pi[codon_i] * row_sum

    for codon_i in codons:
        for codon_j in codons:
            Q[codon_i][codon_j] /= trace

    return Q, trace


def getRateMatrix(trained_model, terminals=None):
    """return a rate matrix from an xrate grammar.

    terminals: return rate matrix and frequencies for these
        terminals. If none are given, a dictionaries of
        matrices and frequencies are returned.


    """

    if terminals:
        xt = (terminals,)
    else:
        xt = trained_model.mGrammar.getTerminals()
        pis, matrices = {}, {}

    for tt in xt:
        # retrieve the terminal frequencies for all codons
        xpi = trained_model.evaluateTerminalFrequencies()[tt]

        pi = {}
        for codon, f in list(xpi.items()):
            pi["".join(codon).upper()] = f

        # retrieve the rate matrix for all codons
        xmatrix = trained_model.evaluateRateMatrix()[tt]
        matrix = {}
        for codon1, v in list(xmatrix.items()):
            x = {}
            for codon2, f in list(xmatrix[codon1].items()):
                x["".join(codon2).upper()] = v
            matrix["".join(codon1).upper()] = x

        if terminals:
            return pi, matrix
        else:
            pis[tt] = pi
            matrices[tt] = matrix

    return pis, matrices


def setFrequencies(model, mali, prefix=""):
    """set frequencies in a model according to those observed in data.

    prefix: prefix for rate parameters.

    Frequencies are labelled:
    pa0, pc0, ..., pa1, pc1, ..., pa2, pc2, ...
    """

    try:
        frequencies = Codons.getFrequenciesPerCodonPosition(
            [x.mString for x in list(mali.values())])
    except XGram.Exceptions.UsageError:
        return

    # build a dummy grammar to insert frequencies
    dummy_grammar = XGram.Model.Grammar()
    for x in range(0, 3):
        params = []
        for a in ('A', 'C', 'G', 'T'):
            params.append(
                ("%sp%s%i" % (prefix, a.lower(), x), frequencies[x][a]))
        dummy_grammar.addVariable(params)

    model.mGrammar.copyParameters(dummy_grammar,
                                  ignore_missing=True)


def getDistanceGTR(pi, matrix):
    """obtain distance from a GTR model.
    see Felsenstein 1994, pp 209
    """
    alphabet = list(pi.keys())
    Q = initializeQMatrix(alphabet)

    # fill matrix AtD and compute trace
    trace = 0
    for i in alphabet:
        row_sum = 0
        for j in alphabet:
            if i == j:
                continue
            v = pi[i] * matrix[i][j]
            Q[i][j] = v
            row_sum += v

        Q[i][i] = -row_sum
        trace += row_sum

    for i in alphabet:
        for j in alphabet:
            Q[i][j] /= trace

    return Q, trace



import random


class Diamond(object):

    def __init__(self):
        self.qid = None
        self.gi = None
        self.ref = None
        self.ngaps = 0
        self.length = 0
        self.evalue = 0
        self.nmismatches = 0
        self.identity = 0
        self.score = 0

    def read(self,
             qid,
             gi,
             ref,
             ngaps,
             length,
             evalue,
             nmismatches,
             identity,
             score):
        '''
        read input
        '''
        self.qid, \
            self.gi, \
            self.ref, \
            self.ngaps, \
            self.length, \
            self.evalue, \
            self.nmismatches, \
            self.identity, \
            self.score = \
            qid, \
            gi, \
            ref, \
            ngaps, \
            length, \
            evalue, \
            nmismatches, \
            identity, \
            score

        return self


def alignment_iterator(alignment_file):
    '''
    return an alignment
    '''
    # replaced .readlines()
    for line in alignment_file:
        data = line[:-1].split("\t")
        ref = data[1].split("|")
        if len(ref) == 1:
            # set ref to reference and leave gi as None
            gi, ref = None, ref[0]
        else:
            gi, ref = ref[1], ref[3]
        read, ngaps, length, evalue, nmismatches, identity, score = data[
            0], data[5], data[3], data[10], data[4], data[2], data[11]
        yield Diamond().read(read,
                             gi,
                             ref,
                             ngaps,
                             length,
                             evalue,
                             nmismatches,
                             identity,
                             score)


def query_iterator(alignment_iterator):
    '''
    iterate over alignments by query
    return a list of alignments
    '''
    reads_found = set()
    last = None
    alignments = []
    for alignment in alignment_iterator:
        this = alignment.qid
        if this != last:
            if last:
                yield alignments
            alignments = []
            reads_found.add(alignment.qid)
        alignments.append(alignment)
        last = this
    if last:
        yield alignments


def best_alignment_iterator(query_iterator):
    '''
    return stats for best alignment - best is determined
    by highest bit score. If there are multiple alignments
    with the same score return a random one
    '''
    for alignments in query_iterator:
        scores = []
        for alignment in alignments:
            scores.append(alignment.score)
        best = max(scores)
        best_alignments = [x for x in alignments if x.score == best]
        if len(best_alignments) > 1:
            best_alignments = random.sample(best_alignments, 1)
        best_alignment = best_alignments[0]
        yield best_alignment



import glob
import sys
import os
import subprocess
import re

########################################################################
#######################################################################
# Check for dependencies
#
# Is there a way to do this more elegantly?
#     1. Run "pip install numpy"
#     2. Wrap inside functions (works for numpy/pysam, but not cython)
try:
    import numpy
except ImportError:
    raise ImportError(
        "the CGAT code collection requires numpy to be installed "
        "before running setup.py (pip install numpy)")

try:
    import Cython
except ImportError:
    raise ImportError(
        "the CGAT code collection requires cython to "
        "be installed before running setup.py (pip install cython)")

try:
    import pysam
except ImportError:
    raise ImportError(
        "the CGAT code collection requires pysam to "
        "be installed before running setup.py (pip install pysam)")

########################################################################
########################################################################
# Import setuptools
# Use existing setuptools, otherwise try ez_setup.
try:
    import setuptools
except ImportError:
    # try to get via ez_setup
    # ez_setup did not work on all machines tested as
    # it uses curl with https protocol, which is not
    # enabled in ScientificLinux
    import ez_setup
    ez_setup.use_setuptools()

from setuptools import setup, find_packages, Extension

from distutils.version import LooseVersion
if LooseVersion(setuptools.__version__) < LooseVersion('1.1'):
    print("Version detected:", LooseVersion(setuptools.__version__))
    raise ImportError(
        "the CGAT code collection requires setuptools 1.1 higher")

from Cython.Distutils import build_ext

########################################################################
########################################################################
IS_OSX = sys.platform == 'darwin'

########################################################################
########################################################################
# collect CGAT version
sys.path.insert(0, "CGAT")
import version

version = version.__version__

###############################################################
###############################################################
# Check for external dependencies
#
# Not exhaustive, simply execute a representative tool from a toolkit.
external_dependencies = (
    ("wigToBigWig", "UCSC tools", 255),
    ("bedtools", "bedtools", 0),
    )

for tool, toolkit, expected in external_dependencies:
    try:
        # py3k
        from subprocess import DEVNULL
    except ImportError:
        DEVNULL = open(os.devnull, 'wb')

    try:
        retcode = subprocess.call(tool, shell=True,
                                  stdout=DEVNULL, stderr=DEVNULL)
    except OSError as msg:
        print(("WARNING: depency check for %s failed: %s" % (toolkit, msg)))

    # UCSC tools return 255 when called without arguments
    if retcode != expected:
        print(("WARNING: depency check for %s(%s) failed, error %i" %
               (toolkit, tool, retcode)))

###############################################################
###############################################################
# Define dependencies
#
major, minor1, minor2, s, tmp = sys.version_info

if (major == 2 and minor1 < 7) or major < 2:
    raise SystemExit("""CGAT requires Python 2.7 or later.""")


#####################################################################
#####################################################################
# Code to install dependencies from a repository
#####################################################################
# Modified from http://stackoverflow.com/a/9125399
#####################################################################
def which(program):
    """
    Detect whether or not a program is installed.
    Thanks to http://stackoverflow.com/a/377028/70191
    """
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ['PATH'].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

REPO_REQUIREMENT = re.compile(
    r'^-e (?P<link>(?P<vcs>git|svn|hg|bzr).+#egg=(?P<package>.+)-(?P<version>\d(?:\.\d)*))$')
HTTPS_REQUIREMENT = re.compile(
    r'^-e (?P<link>.*).+#(?P<package>.+)-(?P<version>\d(?:\.\d)*)$')
install_requires = []
dependency_links = []

for requirement in (
        l.strip() for l in open('requires.txt') if not l.startswith("#")):
    match = REPO_REQUIREMENT.match(requirement)
    if match:
        assert which(match.group('vcs')) is not None, \
            ("VCS '%(vcs)s' must be installed in order to "
             "install %(link)s" % match.groupdict())
        install_requires.append("%(package)s==%(version)s" % match.groupdict())
        dependency_links.append(match.group('link'))
        continue

    if requirement.startswith("https"):
        install_requires.append(requirement)
        continue

    match = HTTPS_REQUIREMENT.match(requirement)
    if match:
        install_requires.append("%(package)s>=%(version)s" % match.groupdict())
        dependency_links.append(match.group('link'))
        continue

    install_requires.append(requirement)

if major == 2:
    install_requires.extend(['web.py>=0.37',
                             'xlwt>=0.7.4',
                             'matplotlib-venn>=0.5'])
elif major == 3:
    pass

cgat_packages = find_packages()
cgat_package_dirs = {'CGAT': 'CGAT'}

##########################################################
##########################################################
# Classifiers
classifiers = """
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: POSIX
Operating System :: Unix
Operating System :: MacOS
"""

##########################################################
# Cython Extensions
# Connected components cython extension
Components = Extension(
    'CGAT.Components',
    ['CGAT/Components/Components.pyx',
     'CGAT/Components/connected_components.cpp', ],
    library_dirs=[],
    libraries=[],
    language="c++",
)

# Nested containment lists
NCL = Extension(
    "CGAT.NCL.cnestedlist",
    ["CGAT/NCL/cnestedlist.pyx",
     "CGAT/NCL/intervaldb.c"],
    library_dirs=[],
    libraries=[],
    language="c",
)

# Timeseries analysis
Timeseries = Extension(
    "CGAT.Timeseries.cmetrics",
    ["CGAT/Timeseries/cmetrics.pyx"],
    include_dirs=[numpy.get_include()],
    library_dirs=[],
    libraries=[],
    language="c",
)

# Nested containment lists
GeneModelAnalysis = Extension(
    "CGAT.GeneModelAnalysis",
    ["CGAT/GeneModelAnalysis.pyx"],
    include_dirs=pysam.get_include() + [numpy.get_include()],
    library_dirs=[],
    libraries=[],
    define_macros=pysam.get_defines(),
    language="c",
)

# automatically build pyximport script extensions
pyx_files = glob.glob("CGAT/scripts/*.pyx")
script_extensions = []
pysam_dirname = os.path.dirname(pysam.__file__)
include_dirs = [numpy.get_include()] + pysam.get_include()

if IS_OSX:
    # linking against bundles does no work (and apparently is not needed)
    # within OS X
    extra_link_args = []
else:
    extra_link_args = [os.path.join(pysam_dirname, x) for x in
                       pysam.get_libraries()]

for pyx_file in pyx_files:
    script_name = os.path.basename(pyx_file)
    script_prefix = script_name[:-4]
    script_extensions.append(
        Extension("CGAT.scripts.%s" % (script_prefix),
                  sources=[pyx_file],
                  extra_link_args=extra_link_args,
                  include_dirs=include_dirs,
                  define_macros=pysam.get_defines())
    )


ext_modules = [Components, NCL, Timeseries, GeneModelAnalysis] + script_extensions

setup(
    # package information
    name='CGAT',
    version=version,
    description='CGAT : the Computational Genomics Analysis Toolkit',
    author='Andreas Heger',
    author_email='andreas.heger@gmail.com',
    license="MIT",
    platforms=["any"],
    keywords="computational genomics",
    long_description='CGAT : the Computational Genomics Analysis Toolkit',
    classifiers=[_f for _f in classifiers.split("\n") if _f],
    url="http://www.cgat.org/cgat/Tools/",
    # package contents
    packages=cgat_packages,
    package_dir=cgat_package_dirs,
    include_package_data=True,
    entry_points={
        'console_scripts': ['cgat = CGAT.cgat:main']
    },
    # dependencies
    install_requires=install_requires,
    dependency_links=dependency_links,
    # extension modules
    ext_modules=ext_modules,
    cmdclass={'build_ext': build_ext},
    # other options
    zip_safe=False,
    test_suite="tests",
)



'''
Variants.py - 
======================================================

:Tags: Python

Code
----

'''
import collections

from CGAT import Experiment as E
from CGAT import Genomics as Genomics
from CGAT import NCL as ncl

Variant = collections.namedtuple("Variant",
                                 "pos, reference, genotype")
ExtendedVariant = collections.namedtuple(
    "Variant",
    "start, end, reference, action, has_wildtype, variantseqs")


def updateVariants(variants, lcontig, strand, phased=True):
    '''update variants such that they use same coordinate
    system (and strand) as the transcript

    fixes 1-ness of variants
    '''

    new_variants = []
    is_positive = Genomics.IsPositiveStrand(strand)

    for variant in variants:

        pos = variant.pos
        genotype = bytes(variant.genotype)
        reference = bytes(variant.reference)

        # fix 1-ness of variants
        # pos -= 1

        if len(genotype) == 1:
            variantseqs = list(Genomics.decodeGenotype(genotype))
            has_wildtype = reference in variantseqs
            action = "="
            start, end = pos, pos + 1
        else:

            variantseqs = [x[1:] for x in genotype.split("/")]
            lvariant = max([len(x) for x in variantseqs])
            if not phased:
                variantseqs = [x for x in variantseqs if x]
            has_wildtype = "*" in genotype

            if "+" in genotype and "-" in genotype:
                # both insertion and deletion at position
                # the range is given by the deletion
                # see below for explanations
                if genotype.startswith("+"):
                    action = ">"
                    variantseqs[1] += "-" * (lvariant - len(variantseqs[1]))
                else:
                    action = "<"
                    variantseqs[0] += "-" * (lvariant - len(variantseqs[0]))

                start, end = pos + 1, pos + lvariant + 1

            elif "-" in genotype:
                action = "-"
                # samtools: deletions are after the base denoted by snp.position
                #   * <- deletion at 1
                # 0 1 2 3 4 5 6
                #     - -
                # 6 5 4 3 2 1 0
                # deletion of 2+3 = (2,4)
                # on reverse: (7-4, 7-2) = (3,5)
                start, end = pos + 1, pos + lvariant + 1

                # deletions of unequal length are filled up with "-"
                # This is necessary to deal with negative strands:
                # -at/-atg on the positive strand deletes a t [g]
                # -at/-atg on the negative strand deletes [g] t a
                variantseqs = [x + "-" * (lvariant - len(x))
                               for x in variantseqs]

            elif "+" in genotype:
                action = "+"
                # indels are after the base denoted by position
                # as region use both flanking base so that negative strand
                # coordinates work
                # insertion between position 2 and 3
                #     * <- insection at pos 2
                # 0 1 2i3 4
                # 4 3 2i1 0
                # is insertion between 1 and 2 in reverse
                # including both flanking residues makes it work:
                # (2,3) = (5-3,5-2) = (2,3)
                # but:
                # (2,4) = (5-4,5-2) = (1,3)
                start, end = pos, pos + 2

        # revert strand
        if not is_positive:
            reference = Genomics.complement(reference)
            variantseqs = [Genomics.complement(x.upper()) for x in variantseqs]
            start, end = lcontig - end, lcontig - start

        new_variants.append(ExtendedVariant._make((
            start, end, reference.upper(), action, has_wildtype, variantseqs)))

    return new_variants


def mergeVariants(variants):
    '''merge overlapping variants.

    Overlapping variants occur if there are two deletions
    at the same location:

        WT      ACTG  
        Allele1 -CT-   
        Allele2 ----

    This will be encoded by samtools as (0-based coordinates)::

        0 * -A/ACTG
        3 * -G/-G

    This upsets the re-constitution algoritm.

    This method separates these two variants into two non-overlapping
    variants making use of variable length deletions.

        0 * -A/-A
        1 * ---G/-CTG

    Another case:

        WT      ACTG  
        Allele1 ACT-   
        Allele2 ----

    This will be encoded by samtools as (0-based coordinates)::

        0 * */-ACTG
        3 * -G/*

    This method separates these two as::

        0 * */-ACT
        3 * -G/-G

    '''

    if len(variants) == 0:
        return []

    # sorts by start and then end
    variants.sort()
    merged_variants = []

    def _add(offset, dest, src):
        for x, c in enumerate(src):
            dest[x + offset] = c

    def _split(seq0, seq1):
        # split
        was_0, was_1 = seq0[0] == "-", seq1[0] == "-"
        for x, cc in enumerate(zip(seq0, seq1)):
            is_0, is_1 = cc[0] == "-", cc[1] == "-"
            # yield all changes
            if (is_0 ^ was_0) or (is_1 ^ was_1):
                yield x, was_0, was_1
            was_0, was_1 = is_0, is_1

        yield x + 1, was_0, was_1

    last = variants[0]
    for this in variants[1:]:

        if this.start < last.end and \
                this.action == "-" and \
                last.action == "-":

            E.warn("merging overlapping deletions: %s and %s" %
                   (str(last), str(this)))

            mend = max(last.end, this.end)
            mstart = min(this.start, last.start)
            l = mend - mstart

            seq0 = list("-" * l)
            seq1 = list("-" * l)

            _add(last.start - mstart, seq0, last.variantseqs[0])
            _add(last.start - mstart, seq1, last.variantseqs[1])
            _add(this.start - mstart, seq0, this.variantseqs[0])
            _add(this.start - mstart, seq1, this.variantseqs[1])

            last_x = 0
            n = []
            for x, was_0, was_1 in _split(seq0, seq1):
                if last_x == x:
                    continue

                this = ExtendedVariant._make((mstart + last_x,
                                              mstart + x,
                                              "*",
                                              last.action,
                                              was_0 ^ was_1,
                                              ["".join(seq0[last_x:x]),
                                               "".join(seq1[last_x:x])],
                                              ))
                n.append(this)
                last_x = x

            E.warn("overlapping deletions merged in %i blocks as: %s" %
                   (len(n), list(map(str, n))))
            merged_variants.extend(n[:-1])
            this = n[-1]
        else:
            merged_variants.append(last)

        last = this

    merged_variants.append(last)

    return merged_variants


def indexVariants(variants):
    '''build index of variants for ranged retrieval.'''
    index = ncl.NCL()

    for start, end, reference, action, has_wildtype, variantseqs in variants:
        index.add(start, end, (reference, action, has_wildtype, variantseqs))

    return index


def buildAlleles(sequence, variants, reference_start=0, phased=True):
    '''build alleles for ``sequence`` adding ``variants``.

    Variants are assumed to be in 0-based coordinates on the same strand as the sequence. 
    ``reference_start`` is the position of the first base of ``sequence``. Set to 0, if
    the positions in ``variants`` are relative to ``sequence``.
    '''

    def _delete(allele, del_start, del_end, variant, sequence, startoffset, endoffset, feature_start, feature_end):
        '''little helper: update ``allele`` with a deletion ``del_start:del_end``.
        '''

        # truncate variant according to the feature
        variant = variant[startoffset:len(variant) - endoffset]

        n = variant.count("-")
        if n:
            if variant.startswith("-"):
                del_start += n
                variant = variant[n:]
            else:
                del_end -= n
                variant = variant[:-n]

        # due to gaps, the variant is not actually within the feauture
        if del_start >= del_end:
            return

        refseq = sequence[del_start:del_end].upper()

        assert refseq == variant, \
            'reference base mismatch at deletion: expected %s %s %s, got %s[%i:%i] at feature=%i-%i, variant=%i-%i, relative=%i-%i, del=%i-%i, action=%s' % \
            (sequence[del_start - 10:del_start],
             refseq,
             sequence[del_end:del_end + 10],
             variant, startoffset, len(variant) - endoffset,
             feature_start, feature_end,
             var_start, var_end,
             rel_start, rel_end,
             del_start, del_end,
             action)

        l = del_end - del_start

        # assert len("".join(allele[del_start:del_end])) == l, \
        #     "deletion conflicts with other indels: " \
        #     "got %s[%i:%i] (ref=%s, allele=%s) at feature=%i-%i, variant=%i-%i, relative=%i-%i, del=%i-%i, action=%s" % \
        #     (variant, startoffset, len(variant)-endoffset,
        #      refseq, str(allele[del_start:del_end]),
        #      feature_start, feature_end,
        #      var_start, var_end,
        #      rel_start, rel_end,
        #      del_start, del_end,
        #      action)

        allele[del_start:del_end] = [""] * l

    allele1 = list(sequence.lower())
    allele2 = list(sequence.lower())

    if reference_start is None:
        feature_start = 0
    else:
        feature_start = reference_start

    feature_end = feature_start + len(sequence)

    # main loop: insert variants into allele sequences
    for var_start, var_end, reference, action, has_wildtype, variantseqs in variants:

        # skip variants that are out-of-range
        if var_end <= feature_start or var_start >= feature_end:
            continue

        is_homozygous = len(variantseqs) == 1 and not has_wildtype

        rel_start, rel_end = var_start - feature_start, var_end - feature_start
        startoffset = max(0, feature_start - var_start)
        endoffset = max(0, var_end - feature_end)
        pruned_start, pruned_end = max(
            0, rel_start), min(len(sequence), rel_end)

        if action == "=":

            if E.global_options.loglevel >= 10:
                E.debug("adding SNP at postition %i: reference=%s variants=%s" %
                        (var_start, reference, variantseqs))

            if allele1[rel_start] == "" or allele2[rel_start] == "":
                # these can be cases, where a base is deleted in one allele,
                # but recorded as a homozygous substitution in another allele.
                E.warn("substitution conflicts with a deletion - ignored: %s" %
                       str((var_start, var_end, reference, action, has_wildtype, variantseqs)))
                continue

            assert rel_start >= 0
            assert sequence[rel_start].upper() == reference, \
                'reference base mismatch: expected %s %s %s, got %s at feature=%i-%i, variant=%i-%i, relative=%i-%i, pruned=%i-%i, action=%s' % \
                (sequence[rel_start - 10:rel_start],
                 sequence[rel_start].upper(),
                 sequence[rel_start + 1:rel_start + 10],
                 reference,
                 feature_start, feature_end,
                 var_start, var_end,
                 rel_start, rel_end,
                 pruned_start, pruned_end,
                 action)

            if phased:
                allele1[rel_start] = variantseqs[0] + allele1[rel_start][1:]
                allele2[rel_start] = variantseqs[1] + allele2[rel_start][1:]
            elif is_homozygous:
                allele1[rel_start] = variantseqs[0] + allele1[rel_start][1:]
                allele2[rel_start] = variantseqs[0] + allele2[rel_start][1:]
            else:
                if has_wildtype:
                    if reference == variantseqs[0]:
                        allele2[rel_start] = variantseqs[
                            1] + allele2[rel_start][1:]
                    else:
                        allele2[rel_start] = variantseqs[
                            0] + allele2[rel_start][1:]
                else:
                    allele1[rel_start] = variantseqs[
                        0] + allele1[rel_start][1:]
                    allele2[rel_start] = variantseqs[
                        1] + allele2[rel_start][1:]

        elif action == "-":
            if phased:
                _delete(allele1, pruned_start, pruned_end, variantseqs[0],
                        sequence, startoffset, endoffset, feature_start, feature_end)
                _delete(allele2, pruned_start, pruned_end, variantseqs[1],
                        sequence, startoffset, endoffset, feature_start, feature_end)
            elif is_homozygous:
                _delete(allele1, pruned_start, pruned_end, variantseqs[0],
                        sequence, startoffset, endoffset, feature_start, feature_end)
                _delete(allele2, pruned_start, pruned_end, variantseqs[0],
                        sequence, startoffset, endoffset, feature_start, feature_end)
            else:
                if has_wildtype:
                    _delete(allele2, pruned_start, pruned_end, variantseqs[0],
                            sequence, startoffset, endoffset, feature_start, feature_end)
                else:
                    _delete(allele1, pruned_start, pruned_end, variantseqs[0],
                            sequence, startoffset, endoffset, feature_start, feature_end)
                    _delete(allele2, pruned_start, pruned_end, variantseqs[1],
                            sequence, startoffset, endoffset, feature_start, feature_end)

        elif action == "+":
            # ignore insertions at position -1
            if rel_start < 0:
                continue

            if phased:
                allele1[rel_start] += variantseqs[0].upper()
                allele2[rel_start] += variantseqs[1].upper()
            elif is_homozygous:
                allele1[rel_start] += variantseqs[0].upper()
                allele2[rel_start] += variantseqs[0].upper()
            else:
                if has_wildtype:
                    allele2[rel_start] += variantseqs[0].upper()
                else:
                    allele1[rel_start] += variantseqs[0].upper()
                    allele2[rel_start] += variantseqs[1].upper()

        elif action == ">":
            # indel
            if rel_start >= 0:
                allele1[rel_start] += variantseqs[0].upper()
            _delete(allele2, pruned_start, pruned_end, variantseqs[1],
                    sequence, startoffset, endoffset, feature_start, feature_end)

        elif action == "<":
            # delin
            if rel_start >= 0:
                allele2[rel_start] += variantseqs[1].upper()
            _delete(allele1, pruned_start, pruned_end, variantseqs[0],
                    sequence, startoffset, endoffset, feature_start, feature_end)

    assert len(sequence) == len(allele1)
    assert len(sequence) == len(allele2)

    return (allele1, allele2)


def buildOffsets(variants, phased=True, contig=None):
    '''collect coordinate offsets.

    This methods takes a set of variants and computes
    coordinates offsets based on indels.

    Conflicting variants will be removed. 

    Returns a list of variants, a list of removed variants and a list of offsets.
    '''

    variants.sort()
    offset0, offset1 = 0, 0
    offsets0, offsets1 = [(0, 0)], [(0, 0)]

    def addDeletion(offsets, start, end, offset, variantseq):

        for x, c in enumerate(variantseq):
            if c != "-":
                offset -= 1
                offsets.append((start + x + 1, offset))

        return offset

    last_end = 0

    removed_variants, kept_variants = [], []
    for variant in variants:

        start, end, reference, action, has_wildtype, variantseqs = variant

        if action == "=":
            kept_variants.append(variant)
            # ignore substitutions
            continue

        if last_end >= start:
            removed_variants.append(variant)
            continue

        kept_variants.append(variant)
        is_homozygous = len(variantseqs) == 1 and not has_wildtype

        last_end = end
        if action == "-":
            if phased:
                offset0 = addDeletion(
                    offsets0, start, end, offset0, variantseqs[0])
                offset1 = addDeletion(
                    offsets1, start, end, offset1, variantseqs[1])
            elif is_homozygous:
                offset0 = addDeletion(
                    offsets0, start, end, offset0, variantseqs[0])
                offset1 = addDeletion(
                    offsets1, start, end, offset1, variantseqs[0])
            else:
                if has_wildtype:
                    offset1 = addDeletion(
                        offsets1, start, end, offset1, variantseqs[0])
                else:
                    offset0 = addDeletion(
                        offsets0, start, end, offset0, variantseqs[0])
                    offset1 = addDeletion(
                        offsets1, start, end, offset1, variantseqs[1])

        elif action == "+":

            if phased:
                offset0 += len(variantseqs[0]) - variantseqs[0].count("-")
                offset1 += len(variantseqs[1]) - variantseqs[1].count("-")
            elif is_homozygous:
                offset0 += len(variantseqs[0])
                offset1 += len(variantseqs[0])
            else:
                if has_wildtype:
                    offset1 += len(variantseqs[0])
                else:
                    offset0 += len(variantseqs[0]) - variantseqs[0].count("-")
                    offset1 += len(variantseqs[1]) - variantseqs[1].count("-")
            offsets0.append((end - 1, offset0))
            offsets1.append((end - 1, offset1))

        elif action == ">":
            offset0 += len(variantseqs[0]) - variantseqs[0].count("-")
            offsets0.append((end - 1, offset0))
            offset1 = addDeletion(
                offsets1, start, end, offset1, variantseqs[1])

        elif action == "<":
            offset0 = addDeletion(
                offsets0, start, end, offset0, variantseqs[0])
            offset1 += len(variantseqs[1]) - variantseqs[1].count("-")
            offsets1.append((end - 1, offset1))

        if E.global_options.loglevel >= 8:
            print("# variant=", start, end, reference,
                  action, has_wildtype, variantseqs)
            print("# offsets:", offsets0[-1], offsets1[-1])
        # add offsets - applies to the end of the variant

    return (kept_variants, removed_variants, (offsets0, offsets1))
